from django.shortcuts import get_object_or_404
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbcore.contrib.directory.models import Person
from wbcore.contrib.directory.serializers import PersonRepresentationSerializer

from wbcompliance.models import (
    ComplianceAction,
    ComplianceEvent,
    ComplianceTask,
    ComplianceTaskGroup,
    ComplianceTaskInstance,
    ComplianceType,
    ReviewComplianceTask,
)

from .compliance_type import ComplianceTypeRepresentationSerializer


class ComplianceTaskGroupRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcompliance:compliancetaskgroup-detail")

    class Meta:
        model = ComplianceTaskGroup
        fields = ("id", "name", "_detail")


class ComplianceTaskRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcompliance:compliancetask-detail")

    class Meta:
        model = ComplianceTask
        fields = ("id", "title", "occurrence", "active", "type", "_detail")


class ReviewComplianceTaskRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcompliance:reviewcompliancetask-detail")

    class Meta:
        model = ReviewComplianceTask
        fields = ("id", "title", "computed_str", "occurrence", "year", "occured", "_detail")


class NoInstanceReviewComplianceTaskRepresentationSerializer(ReviewComplianceTaskRepresentationSerializer):
    def get_filter_params(self, request):
        if task_id := request.parser_context["view"].kwargs.get("pk"):
            task = get_object_or_404(ComplianceTask, pk=task_id)
            return {"is_instance": False, "type": task.type.id}
        return {"is_instance": False}


class InstanceReviewComplianceTaskRepresentationSerializer(ReviewComplianceTaskRepresentationSerializer):
    def get_filter_params(self, request):
        if task_id := request.parser_context["view"].kwargs.get("task_id"):
            task = get_object_or_404(ComplianceTask, pk=task_id)
            return {"is_instance": True, "type": task.type.id}
        return {"is_instance": True}


class ComplianceTaskGroupModelSerializer(wb_serializers.ModelSerializer):
    @wb_serializers.register_resource()
    def register_history_resource(self, instance, request, user):
        resources = {
            "compliancetask": reverse(
                "wbcompliance:compliancetaskgroup-compliancetask-list", args=[instance.id], request=request
            ),
        }
        return resources

    class Meta:
        model = ComplianceTaskGroup
        fields = ("id", "name", "order", "_additional_resources")


class ComplianceTaskModelSerializer(wb_serializers.ModelSerializer):
    @wb_serializers.register_resource()
    def register_history_resource(self, instance, request, user):
        resources = {
            "compliancetaskinstance": reverse(
                "wbcompliance:compliancetask-compliancetaskinstance-list", args=[instance.id], request=request
            ),
        }
        if user.is_superuser:
            resources["generateinstance"] = reverse(
                "wbcompliance:compliancetask-generate-instance", args=[instance.id], request=request
            )
        return resources

    _group = ComplianceTaskGroupRepresentationSerializer(source="group")
    review = wb_serializers.PrimaryKeyRelatedField(
        queryset=ReviewComplianceTask.objects.all(),
        label=_("Indicator Reports"),
        many=True,
        help_text=_("list of reports that contain this task"),
    )
    _review = NoInstanceReviewComplianceTaskRepresentationSerializer(
        source="review",
        many=True,
        optional_get_parameters={"type": "type"},
        depends_on=[{"field": "type", "options": {}}],
    )
    _type = ComplianceTypeRepresentationSerializer(source="type")
    type = wb_serializers.PrimaryKeyRelatedField(
        queryset=ComplianceType.objects.all(),
        label=_("Administrator"),
    )
    occurrence = wb_serializers.ChoiceField(
        default=wb_serializers.DefaultAttributeFromRemoteField("review_id", ReviewComplianceTask, "occurrence"),
        choices=ReviewComplianceTask.Occurrence.choices,
    )
    type = wb_serializers.PrimaryKeyRelatedField(
        queryset=ComplianceType.objects.all(),
        label=_("Administrator"),
        default=wb_serializers.DefaultAttributeFromRemoteField("review_id", ReviewComplianceTask, "type.id"),
    )

    class Meta:
        model = ComplianceTask
        # dependency_map = {
        #     "review": ["type"],
        # }
        fields = (
            "id",
            "title",
            "description",
            "occurrence",
            "active",
            "type",
            "_type",
            "group",
            "_group",
            "_additional_resources",
            "risk_level",
            "remarks",
            "review",
            "_review",
        )


class ComplianceTaskReviewModelSerializer(ComplianceTaskModelSerializer):
    def create(self, validated_data):
        if (view := self.context.get("view", None)) and (review_id := view.kwargs.get("review_id", None)):
            validated_data["review"] = ReviewComplianceTask.objects.filter(id=review_id)

        return super().create(validated_data)


class ComplianceTaskInstanceModelSerializer(wb_serializers.ModelSerializer):
    task = wb_serializers.PrimaryKeyRelatedField(
        queryset=ComplianceTask.objects.all(),
        label=_("Indicator"),
    )
    _task = ComplianceTaskRepresentationSerializer(source="task")
    type_name = wb_serializers.CharField(required=False, read_only=True, label=_("Administrator"))
    group_name = wb_serializers.CharField(required=False, read_only=True, label=_("Group"))
    task_title = wb_serializers.CharField(required=False, read_only=True, label=_("Indicators"))
    review = wb_serializers.PrimaryKeyRelatedField(
        queryset=ReviewComplianceTask.objects.all(),
        label=_("Indicator Instance Reports"),
        many=True,
        help_text=_("list of reports that contain this task"),
    )
    _review = InstanceReviewComplianceTaskRepresentationSerializer(source="review", many=True)

    text = wb_serializers.TextAreaField(
        required=False, label=_("Text"), allow_blank=True, allow_null=True, help_text=_("Information for Management")
    )
    summary_text = wb_serializers.TextAreaField(
        required=False,
        label=_("Summary Text"),
        allow_blank=True,
        allow_null=True,
        help_text=_("Information for the Board of Directors"),
    )

    class Meta:
        model = ComplianceTaskInstance
        fields = (
            "id",
            "occured",
            "status",
            "type_name",
            "group_name",
            "text",
            "summary_text",
            "_task",
            "task",
            "review",
            "_review",
            "_additional_resources",
            "task_title",
        )
        read_only_fields = ("review", "_review", "task", "_task")


class ComplianceTaskInstanceListModelSerializer(ComplianceTaskInstanceModelSerializer):
    class Meta:
        model = ComplianceTaskInstance
        fields = (
            "id",
            "occured",
            "status",
            "type_name",
            "group_name",
            "_task",
            "task",
            "task_title",
            "review",
            "_review",
        )


class ComplianceActionModelSerializer(wb_serializers.ModelSerializer):
    type = wb_serializers.PrimaryKeyRelatedField(
        queryset=ComplianceType.objects.all(),
        label=_("Administrator"),
    )
    _type = ComplianceTypeRepresentationSerializer(source="type")
    _creator = PersonRepresentationSerializer(source="creator")
    _changer = PersonRepresentationSerializer(source="changer")
    description = wb_serializers.TextField(
        default="", label=_("Description"), help_text=_("Explanation for Management")
    )
    summary_description = wb_serializers.TextField(
        default="", label=_("Summary Description"), help_text=_("Explanation for the Board of Directors")
    )

    class Meta:
        percent_fields = ["progress"]

        model = ComplianceAction
        fields = (
            "id",
            "title",
            "description",
            "summary_description",
            "deadline",
            "progress",
            "status",
            "type",
            "_type",
            "active",
            "creator",
            "_creator",
            "created",
            "changer",
            "_changer",
            "last_modified",
        )
        read_only_fields = ("creator", "created", "changer", "last_modified")

    @cached_property
    def user_profile(self) -> Person | None:
        if request := self.context.get("request"):
            return request.user.profile
        return None

    def create(self, validated_data):
        validated_data["creator"] = self.user_profile
        return super().create(validated_data)

    def update(self, instance, validated_data):
        profile = self.user_profile
        validated_data["changer"] = profile
        validated_data.pop("creator", None)
        validated_data.pop("created", None)
        return super().update(instance, validated_data)


class ComplianceEventModelSerializer(wb_serializers.ModelSerializer):
    type = wb_serializers.PrimaryKeyRelatedField(
        queryset=ComplianceType.objects.all(),
        label=_("Administrator"),
    )
    _type = ComplianceTypeRepresentationSerializer(source="type")
    _creator = PersonRepresentationSerializer(source="creator")
    _changer = PersonRepresentationSerializer(source="changer")
    title = wb_serializers.CharField(label=_("Title"), allow_null=False)
    exec_summary = wb_serializers.TextField(
        default="", label=_("Executive Summary for Management"), help_text=_("Executive Summary for Management")
    )
    exec_summary_board = wb_serializers.TextField(
        default="",
        label=_("Executive Summary for the Board of Directors"),
        help_text=_("Executive Summary for the Board of Directors"),
    )

    class Meta:
        model = ComplianceEvent
        fields = (
            "id",
            "type_event",
            "level",
            "title",
            "exec_summary",
            "exec_summary_board",
            "description",
            "actions_taken",
            "confidential",
            "consequences",
            "future_suggestions",
            "type",
            "_type",
            "active",
            "creator",
            "_creator",
            "created",
            "changer",
            "_changer",
            "last_modified",
        )
        read_only_fields = ("creator", "created", "changer", "last_modified")

    @cached_property
    def user_profile(self) -> Person | None:
        if request := self.context.get("request"):
            return request.user.profile
        return None

    def create(self, validated_data):
        validated_data["creator"] = self.user_profile
        return super().create(validated_data)

    def update(self, instance, validated_data):
        profile = self.user_profile
        validated_data["changer"] = profile
        validated_data.pop("creator", None)
        validated_data.pop("created", None)
        return super().update(instance, validated_data)


class ReviewComplianceTaskModelSerializer(wb_serializers.ModelSerializer):
    @wb_serializers.register_only_instance_resource()
    def register_history_resource(self, instance, request, user, **kwargs):
        if not instance.is_instance:
            resources = {
                "task_group": reverse("wbcompliance:compliancetaskgroup-list", args=[], request=request),
                "task_no_group": reverse(
                    "wbcompliance:review-compliancetasknogroup-list", args=[instance.id], request=request
                ),
            }
            if instance.status == ReviewComplianceTask.Status.DRAFT and user.is_superuser:
                resources["link_tasks"] = reverse(
                    "wbcompliance:reviewcompliancetask-link-tasks", args=[instance.id], request=request
                )

            if instance.status == ReviewComplianceTask.Status.VALIDATED and user.has_perm(
                "wbcompliance.administrate_compliance"
            ):
                resources["regenerate_document"] = reverse(
                    "wbcompliance:reviewcompliancetask-regenerate-document", args=[instance.id], request=request
                )

                from_date, to_date = instance.get_period_date()
                if not ReviewComplianceTask.objects.filter(
                    review_task=instance, from_date=from_date, to_date=to_date
                ).exists():
                    resources["generate_instance"] = reverse(
                        "wbcompliance:reviewcompliancetask-generate-instance", args=[instance.id], request=request
                    )
            group_ids = instance.get_task_group_ids_from_review()
            for group_id in group_ids:
                resources[f"taskgroup{group_id}"] = reverse(
                    "wbcompliance:review-compliancetaskgroup-list",
                    args=[instance.id, group_id],
                    request=request,
                )
        else:
            resources = {
                "actions": reverse(
                    "wbcompliance:type-complianceaction-list", args=[instance.type.id], request=request
                ),
                "events": reverse("wbcompliance:type-complianceevent-list", args=[instance.type.id], request=request),
                "taskinstance_no_group": reverse(
                    "wbcompliance:review-compliancetaskinstancenogroup-list", args=[instance.id], request=request
                ),
            }
            group_ids = instance.get_task_group_ids_from_review(through_task=False)
            for group_id in group_ids:
                resources[f"taskinstancegroup{group_id}"] = reverse(
                    "wbcompliance:review-compliancetaskinstancegroup-list",
                    args=[instance.id, group_id],
                    request=request,
                )

        return resources

    year = wb_serializers.YearField(default=timezone.now().year)
    creator = wb_serializers.PrimaryKeyRelatedField(
        default=wb_serializers.CurrentUserDefault("profile"),
        many=False,
        read_only=True,
    )
    _creator = PersonRepresentationSerializer(source="creator")
    _changer = PersonRepresentationSerializer(source="changer")
    review_task = wb_serializers.PrimaryKeyRelatedField(read_only=True, label=_("Main Indicator Report"))
    _review_task = ReviewComplianceTaskRepresentationSerializer(source="review_task")
    type = wb_serializers.PrimaryKeyRelatedField(
        queryset=ComplianceType.objects.all(),
        label=_("Administrator"),
    )
    _type = ComplianceTypeRepresentationSerializer(source="type")
    occured = wb_serializers.DateField(read_only=True, label=_("Occured Instance Report"))
    occurrence = wb_serializers.ChoiceField(
        choices=ReviewComplianceTask.Occurrence.choices,
        read_only=lambda view: not getattr(view.get_object(), "is_instance", True)
        if "pk" in view.kwargs
        else not view.new_mode,
    )

    class Meta:
        model = ReviewComplianceTask
        fields = (
            "id",
            "title",
            "from_date",
            "to_date",
            "description",
            "year",
            "creator",
            "_creator",
            "created",
            "changer",
            "_changer",
            "changed",
            "status",
            "_additional_resources",
            "occurrence",
            "computed_str",
            "is_instance",
            "review_task",
            "_review_task",
            "occured",
            "type",
            "_type",
        )
        read_only_fields = (
            "creator",
            "created",
            "changer",
            "changed",
            "occured",
            "is_instance",
            "review_task",
        )

    @cached_property
    def user_profile(self) -> Person | None:
        if request := self.context.get("request"):
            return request.user.profile
        return None

    def create(self, validated_data):
        validated_data["creator"] = self.user_profile
        return super().create(validated_data)

    def update(self, instance, validated_data):
        validated_data["changer"] = self.user_profile
        validated_data.pop("creator", None)
        validated_data.pop("created", None)
        return super().update(instance, validated_data)



from .compliance_form import (
    ComplianceFormAdmin,
    ComplianceFormRuleAdmin,
    ComplianceFormSectionAdmin,
    ComplianceFormSignatureAdmin,
    ComplianceFormSignatureRuleAdmin,
    ComplianceFormSignatureSectionAdmin,
    ComplianceFormTypeAdmin,
)
from .compliance_task import (
    ComplianceActionAdmin,
    ComplianceEventAdmin,
    ReviewComplianceTaskAdmin,
)
from .compliance_type import ComplianceTypeAdmin
from .risk_management import *



from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.utils.translation import gettext as _
from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)

from wbcompliance.models import ComplianceType, ReviewComplianceTask
from wbcompliance.viewsets.buttons.compliance_form import (
    AbstractComplianceDocumentButtonConfig,
)


def get_custom_report(_request, review: ReviewComplianceTask):
    try:
        content_type = ContentType.objects.get(app_label="wbcompliance", model="reviewcompliancetask")
        ReportVersion = apps.get_model("wbreport", "ReportVersion")
        BUTTTONS = []
        versions = ReportVersion.objects.filter(
            version_date=review.to_date, report__content_type=content_type, report__is_active=True
        ).filter(Q(parameters__type=review.type.name) | Q(parameters__type=review.type.id))
        for version in versions:
            BUTTTONS.append(
                bt.HyperlinkButton(
                    endpoint=reverse("report:reportversion-html", args=[version.id], request=_request),
                    label=_("Show {}").format(version.title),
                    icon=WBIcon.DOCUMENT.icon,
                )
            )
            if not version.lock:
                BUTTTONS.append(
                    bt.ActionButton(
                        method=RequestType.GET,
                        identifiers=("wbreport:reportversion",),
                        endpoint=reverse("wbreport:reportversion-updatecontext", args=[version.id], request=_request),
                        label=_("Update {}").format(version.title),
                        description_fields=_(
                            """
                        <p>Update and actualize context</p>
                        """
                        ),
                        icon=WBIcon.REGENERATE.icon,
                        action_label=_("Update Context {}").format(version.title),
                        title=_("Update {}").format(version.title),
                    ),
                )
        if BUTTTONS:
            return {
                bt.DropDownButton(
                    label=_("Report"),
                    icon=WBIcon.UNFOLD.icon,
                    buttons=tuple(BUTTTONS),
                ),
            }
    except LookupError:
        pass

    return set()


class ComplianceTaskButtonConfig(ButtonViewConfig):
    def get_custom_instance_buttons(self):
        class OccuredSerializer(wb_serializers.Serializer):
            occured = wb_serializers.DateField(required=True)

        BUTTONS = []
        if self.view.kwargs.get("pk", None) and self.request.user.is_superuser:
            review = self.view.get_object()
            if review.active:
                BUTTONS += [
                    bt.ActionButton(
                        method=RequestType.PATCH,
                        identifiers=("wbcompliance:compliancetask",),
                        key="generateinstance",
                        action_label=_("Generate Instance"),
                        title=_("Generate Instance "),
                        label=_("Generate Instance"),
                        icon=WBIcon.ADD.icon,
                        serializer=OccuredSerializer,
                        instance_display=create_simple_display([["occured"]]),
                    )
                ]
        return {*BUTTONS}


class ReviewComplianceTaskButtonConfig(ButtonViewConfig):
    def get_custom_instance_buttons(self):
        class LinkedTypeSerializer(wb_serializers.Serializer):
            type = wb_serializers.ChoiceField(
                required=True, choices=ComplianceType.objects.all().values_list("name", flat=True)
            )
            operation = wb_serializers.ChoiceField(
                choices=[("ADD", "Add existing indicators"), ("REMOVE", "Remove existing Indicators")],
            )

        BUTTONS = []
        if self.view.kwargs.get("pk", None) and self.request.user.has_perm("wbcompliance.administrate_compliance"):
            review = self.view.get_object()
            if review.is_instance:
                BUTTONS = get_custom_report(_request=self.request, review=review)
            else:
                if review.status == ReviewComplianceTask.Status.VALIDATED:
                    BUTTONS += [AbstractComplianceDocumentButtonConfig._get_compliance_document_button()]

                    if self.request.user.has_perm("wbcompliance.administrate_compliance"):
                        BUTTONS += [
                            bt.ActionButton(
                                method=RequestType.PATCH,
                                identifiers=("wbcompliance:reviewcompliancetask",),
                                key="generate_instance",
                                action_label="Generate Instance Report Indicator",
                                title="Generate Instance Report Indicator",
                                label="Generate Instance Report Indicator",
                                icon=WBIcon.ADD.icon,
                                confirm_config=bt.ButtonConfig(label=_("Confirm")),
                                cancel_config=bt.ButtonConfig(label=_("Cancel")),
                            ),
                        ]

                if review.status == ReviewComplianceTask.Status.DRAFT and self.request.user.is_superuser:
                    BUTTONS += [
                        bt.ActionButton(
                            method=RequestType.PATCH,
                            identifiers=("wbcompliance:reviewcompliancetask",),
                            key="link_tasks",
                            action_label=_("Add/Remove Indicators of a Type"),
                            title=_("Add/Remove existing indicators"),
                            label=_("Add/Remove existing indicators"),
                            icon=WBIcon.LINK.icon,
                            description_fields=_(
                                """
                            <p> <b>Indicators Report : <span>{{title}} </span></b><br/><br/>
                            Linking Indicators of type
                            """
                            ),
                            serializer=LinkedTypeSerializer,
                            instance_display=create_simple_display([["type"], ["operation"]]),
                            confirm_config=bt.ButtonConfig(label=_("Confirm")),
                            cancel_config=bt.ButtonConfig(label=_("Cancel")),
                        ),
                    ]

        return {*BUTTONS}



from .checks import RiskCheckFactory
from .incidents import (
    CheckedObjectIncidentRelationshipFactory,
    RiskIncidentFactory,
    RiskIncidentTypeFactory,
)
from .rules import (
    RiskRuleFactory,
    RuleBackendFactory,
    RuleCheckedObjectRelationshipFactory,
    RuleThresholdFactory,
)



from contextlib import suppress
from datetime import date, datetime, timedelta
from importlib import import_module
from typing import Any, Dict, Generator, Iterable, Iterator, Optional, Type

import pandas as pd
from celery import shared_task
from dateutil import rrule
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.contrib.auth.models import User as BaseUser
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.fields import DecimalRangeField
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.db.models import Exists, OuterRef, Q
from django.db.models.signals import pre_delete, pre_save
from django.dispatch import receiver
from django.template.loader import get_template
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from pandas.tseries.offsets import BDay
from psycopg.types.range import DateRange
from rest_framework.reverse import reverse
from wbcore.contrib.directory.models import Person
from wbcore.contrib.guardian.models.mixins import PermissionObjectModelMixin
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.models import WBModel
from wbcore.utils.models import ComplexToStringMixin
from wbcore.utils.rrules import convert_rrulestr_to_dict

from .backend import AbstractRuleBackend
from .checks import RiskCheck
from .incidents import CheckedObjectIncidentRelationship, RiskIncident, RiskIncidentType

User: BaseUser = get_user_model()


class RuleGroup(models.Model):
    key = models.CharField(max_length=255, unique=True)
    name = models.CharField(max_length=255)

    def save(self, *args, **kwargs):
        if not self.name:
            self.name = self.key.title()
        super().save(*args, **kwargs)

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcompliance:rulegrouprepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{name}}"


class RuleCheckedObjectRelationship(ComplexToStringMixin, models.Model):
    rule = models.ForeignKey(
        to="wbcompliance.RiskRule", related_name="checked_object_relationships", on_delete=models.CASCADE
    )
    checked_object_content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, related_name="risk_management_checked_objects"
    )
    checked_object_id = models.PositiveIntegerField()
    checked_object = GenericForeignKey("checked_object_content_type", "checked_object_id")
    checked_object_repr = models.CharField(max_length=256, blank=True, null=True)

    class Meta:
        verbose_name = "Checked Object to Rule relationship"
        verbose_name_plural = "Checked Object to Rule relationships"
        indexes = [
            models.Index(fields=["rule", "checked_object_content_type", "checked_object_id"]),
        ]
        unique_together = [("rule", "checked_object_content_type", "checked_object_id")]

    @property
    def checks(self) -> "models.QuerySet[RiskCheck]":
        return RiskCheck.objects.filter(
            checked_object_id=self.checked_object_id,
            checked_object_content_type=self.checked_object_content_type,
            rule=self.rule,
        )

    def clean(self):
        allowed_checked_object_content_type = self.rule.rule_backend.allowed_checked_object_content_type
        if allowed_checked_object_content_type and (
            allowed_checked_object_content_type != self.checked_object_content_type
        ):
            raise ValidationError(
                _(
                    "The relationship content type ({}) needs to match the rule's backend allowed content type ({})"
                ).format(self.checked_object_content_type, allowed_checked_object_content_type)
            )
        super().clean()

    def save(self, *args, **kwargs):
        self.full_clean()
        self.checked_object_repr = str(self.checked_object)
        super().save(*args, **kwargs)

    def compute_str(self) -> str:
        return _("Rule Relationship {} -> {} {}").format(
            self.rule, self.checked_object_content_type.name, self.checked_object
        )

    def process_rule(self, evaluation_date: date, override_incident: bool = False) -> bool:
        """
        Trigger the check between the rule and the checked object attached to this relationship
        """
        rule_backend = self.rule.rule_backend.backend(
            evaluation_date, self.checked_object, self.rule.parameters, self.rule.thresholds.all()
        )
        incident_detected = False
        if rule_backend.is_passive_evaluation_valid():
            check = RiskCheck.objects.create(
                rule=self.rule,
                checked_object_id=self.checked_object_id,
                checked_object_content_type=self.checked_object_content_type,
                evaluation_date=evaluation_date,
            )
            # we create the check but the rule might not be allowed to be processed on that particular date (e.g. wrong frequency)
            if self.rule.is_evaluation_date_valid(evaluation_date):
                for incident in check.evaluate(override_incident=override_incident):
                    incident_detected = True
                    incident.post_workflow()
        return incident_detected

    def get_unchecked_dates(
        self, from_date: Optional[date] = None, to_date: Optional[date] = None, maximum_day_interval: int = 30
    ) -> Iterator[date]:
        """
        if a checks exists it generates all dates between it and the specified to_date (if exists, otherwise, return just the next day after the last check
        if checks does not exist but to_date is specified, generates a unique date

        Args:
            to_date: The limit at which we want to check the next expected check date. Defaults None (no upper bound).

        Returns:
            A list of date to be checked
        """

        if not from_date:
            if self.checks.exists():
                from_date = (self.checks.latest("evaluation_date").evaluation_date + BDay(1)).date()
            elif to_date:
                from_date = to_date - timedelta(days=maximum_day_interval)
            if to_date:
                minimum_allowed_from_date = to_date - timedelta(days=maximum_day_interval)
                from_date = max([from_date, minimum_allowed_from_date])
        if not to_date:
            to_date = from_date
        if not from_date and not to_date:
            raise ValueError("Either from or To date needs to be provided")
        for evaluation_date in pd.date_range(from_date, to_date, freq="B"):
            if not self.checks.filter(evaluation_date=evaluation_date.date()).exists():
                yield evaluation_date.date()

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcompliance:rulechecked_objectrelationshiprepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"


class RuleBackend(models.Model):
    """
    Represent a rule backend  that links to a module where a RuleBackend class is defined.

    We expect this class to at least define the following interface:
    * check(evaluation_date, checked_object, **json_parameters) // Check for a given date, triggering object, a set of
    parameters and corresponding severity thresholds if a rule is in breach.

    Optionally, it can define the following methods:
    * is_passive_evaluation_valid(eval_date, evaluated_object)
    """

    name = models.CharField(max_length=128)
    backend_class_path = models.CharField(max_length=512)
    backend_class_name = models.CharField(max_length=128, default="RuleBackend")
    allowed_checked_object_content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, blank=True, null=True
    )
    rule_group = models.ForeignKey(
        to="wbcompliance.RuleGroup", related_name="rules", null=True, blank=True, on_delete=models.SET_NULL
    )
    incident_report_template = models.TextField(
        default=get_template("risk_management/incident_report.html").template.source, verbose_name="Incident Template"
    )

    class Meta:
        verbose_name = "Rule Backend"
        verbose_name_plural = "Rule Backends"

    @cached_property
    def backend_class(self) -> type[AbstractRuleBackend]:
        """
        Return the imported backend class
        Returns:
            The backend class
        """
        return getattr(import_module(self.backend_class_path), self.backend_class_name)

    def backend(
        self,
        evaluation_date: date,
        evaluated_object: models.Model,
        json_parameters: Dict[str, Any],
        thresholds: "models.QuerySet[RuleThreshold]",
    ) -> AbstractRuleBackend:
        """
        Args:
            evaluation_date: The evaluation rule date
            evaluated_object: The object that needs evaluation
            json_parameters: Set of paramaters as dictionary. Might expect deserialization that will be handled within the backend
            thresholds: List of numerical range, severity pairs

        Return the instantiated backend imported through the specified dotted path and the passed parameters
        Returns:
            The instantiated backend
        """
        checked_object_content_type = ContentType.objects.get_for_model(evaluated_object)
        if self.allowed_checked_object_content_type and (
            checked_object_content_type != self.allowed_checked_object_content_type
        ):
            raise ValidationError(
                _("Passed content type {} does not match the allowed backend content type {}").format(
                    checked_object_content_type, self.allowed_checked_object_content_type
                )
            )
        return self.backend_class(
            evaluation_date, evaluated_object, json_parameters, thresholds.order_by("severity__severity_order")
        )

    def get_all_active_relationships(self) -> Iterable:
        try:
            return self.backend_class.get_all_active_relationships()
        except NotImplementedError:
            return []

    def __str__(self):
        return _("Rule Backend {}").format(self.name)

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcompliance:rulebackendrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{name}}"


class RuleThreshold(ComplexToStringMixin, models.Model):
    """
    Represent the list of threshold and its associated severity link to a certain rule
    """

    rule = models.ForeignKey(to="wbcompliance.RiskRule", related_name="thresholds", on_delete=models.CASCADE)
    range = DecimalRangeField(
        verbose_name=_("Threshold range"),
        help_text=_("The range which triggers the specified severity. null bound represent infinity"),
    )

    severity = models.ForeignKey(
        "wbcompliance.RiskIncidentType",
        on_delete=models.CASCADE,
        verbose_name=_("Triggered Severity"),
        help_text=_("The Triggered Severity when the rule is within the threshold range"),
        related_name="thresholds",
    )
    upgradable_after_days = models.PositiveIntegerField(
        blank=True,
        null=True,
        verbose_name=_("Upgradable to next severity after X Days"),
        help_text=_(
            "If set to a positive integer, the resulting incident will be elevated to the next rule severity after an incident remains open this number of days"
        ),
    )

    notifiable_users = models.ManyToManyField(
        "directory.Person",
        related_name="notified_rule_thresholds",
        blank=True,
        verbose_name=_("Notifiable Persons"),
        help_text=_("Notified Persons for this rule and severity"),
    )
    notifiable_groups = models.ManyToManyField(
        Group,
        related_name="notified_rule_thresholds",
        blank=True,
        verbose_name=_("Notifiable Groups"),
        help_text=_("Notified Groups for this rule and severity"),
    )

    @property
    def next_threshold(self):
        """
        Property that hold the next higher severity order threshold

        Returns: The next Threshold in the severity order sense
        """
        higher_thresholds = (
            self.rule.thresholds.exclude(id=self.id)
            .filter(severity__severity_order__gt=self.severity.severity_order)
            .order_by("severity__severity_order")
        )

        if higher_thresholds.exists():
            return higher_thresholds.first()

    @property
    def numerical_range(self) -> tuple[float, float]:
        return float(self.range.lower) if self.range.lower is not None else float("-inf"), float(  # type: ignore
            self.range.upper  # type: ignore
        ) if self.range.upper is not None else float("inf")  # type: ignore

    @property
    def range_repr(self) -> str:
        upper_bound_repr = "{}]".format(self.range.upper) if self.range.upper is not None else "∞["  # type: ignore
        lower_bound_repr = "[{}".format(self.range.lower) if self.range.lower is not None else "]-∞"  # type: ignore
        return f"{lower_bound_repr}, {upper_bound_repr}"

    def is_inrange(self, value) -> bool:
        """
        Utility function to check whether is within this threshold range
        Args:
            value: The value to check against

        Returns:
            True if the value is in range
        """
        return value > self.numerical_range[0] and value < self.numerical_range[1]

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)

    def compute_str(self) -> str:
        return _("Range: {} (Severity {})").format(self.range_repr, self.severity)

    def get_notifiable_users(self) -> "models.QuerySet[Person]":
        """
        Returns the incident notifiable persons

        Returns:
            A queryset of Person
        """
        try:
            return User.objects.filter(
                models.Q(profile__in=self.notifiable_users.values("id"))
                | models.Q(groups__in=self.notifiable_groups.all())
            ).distinct()
        except ObjectDoesNotExist:
            return User.objects.none()

    class Meta:
        unique_together = ("rule", "severity")

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcompliance:rulethresholdrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"


class RiskRuleDefaultManager(models.Manager):
    def get_active_rules_for_object(self, obj: Type[models.Model]) -> models.QuerySet["RiskRule"]:
        return (
            self.get_queryset()
            .annotate(
                has_direct_relationship=Exists(
                    RuleCheckedObjectRelationship.objects.filter(
                        rule=OuterRef("pk"),
                        checked_object_content_type=ContentType.objects.get_for_model(obj),
                        checked_object_id=obj.id,
                    )
                )
            )
            .filter(
                Q(is_enable=True)
                & Q(only_passive_check_allowed=False)
                & (Q(apply_to_all_active_relationships=True) | Q(has_direct_relationship=True))
            )
        )


class RiskRule(PermissionObjectModelMixin, WBModel):
    """
    Base class for rule management. All the data that defined the rule parameters.

    Expected to be inherited (Abstract)
    """

    name = models.CharField(max_length=512)
    description = models.CharField(max_length=516, default="", verbose_name=_("Quick Description"))
    rule_backend = models.ForeignKey(to="wbcompliance.RuleBackend", related_name="rules", on_delete=models.CASCADE)

    is_enable = models.BooleanField(default=True, verbose_name=_("Enabled"))
    only_passive_check_allowed = models.BooleanField(
        default=True,
        verbose_name=_("Passive Only"),
        help_text=_("If False, This rule can only be triggered passively"),
    )
    is_silent = models.BooleanField(
        default=True, help_text=_("If true, the notification won't be send through System nor Mail")
    )
    is_mandatory = models.BooleanField(
        default=False, help_text=_("A mandatory rule cannot be modified by anyone other than the administrators")
    )
    automatically_close_incident = models.BooleanField(
        default=False, help_text=_("If True, this rule will automatically close all encountered incidents")
    )

    apply_to_all_active_relationships = models.BooleanField(
        default=False,
        help_text=_("If True, will keep this rule syncrhonize with the backend definition of all active relationship"),
    )
    activation_date = models.DateField(null=True, blank=True, verbose_name="Activation Date")
    frequency = models.CharField(
        null=True,
        blank=True,
        max_length=56,
        verbose_name=_("Evaluation Frequency"),
        help_text=_("The Evaluation Frequency in RRULE format"),
    )
    parameters = models.JSONField(blank=True, default=dict, encoder=DjangoJSONEncoder)

    class Meta(PermissionObjectModelMixin.Meta):
        verbose_name = "Risk Rule"
        verbose_name_plural = "Risk Rules"

    objects = RiskRuleDefaultManager()

    @property
    def checked_object_representation(self) -> str:
        try:
            backend_class = self.rule_backend.backend_class
            checked_object_repr = getattr(backend_class, "OBJECT_FIELD_NAME").title()
        except AttributeError:
            checked_object_repr = "Object"
        return checked_object_repr

    @property
    def checked_objects(self) -> Generator[models.Model, None, None]:
        """
        All objects that share a relationship with this rule instance

        Returns:
            An generator of Object (any type)

        """
        for relation in self.checked_object_relationships.all():
            yield relation.checked_object

    @property
    def checks(self) -> "models.QuerySet[RiskCheck]":
        return RiskCheck.objects.filter(rule=self).distinct()

    def save(self, *args, **kwargs):
        serializer = self.rule_backend.backend_class.get_serializer_class()(data=self.parameters)
        if serializer.is_valid():
            self.parameters = serializer.data
        super().save(*args, **kwargs)

    def __str__(self):
        return _("Rule {}").format(self.name)

    def is_evaluation_date_valid(self, evaluation_date: date) -> bool:
        if self.frequency and self.activation_date:
            rrule_dict = convert_rrulestr_to_dict(self.frequency, dtstart=self.activation_date, until=evaluation_date)
            valid_dates = list(map(lambda o: o.date(), rrule.rrule(**rrule_dict)))
            return evaluation_date in valid_dates
        return self.activation_date is not None and self.activation_date <= evaluation_date

    def process_rule(self, evaluation_date: date, override_incident: bool = False, silent_notification: bool = False):
        """
        Wrapper function that calls the not implemented check_risk to evaluate the rule against a date and all linked risk management instances.

        Args:
            evaluation_date: The evaluation date
            override_incident: If True, the existing incidents will be reopened instead of being ignored
            silent_notification: If True, explicitly silent notification even if they are due
        """
        if not self.is_enable:
            raise ValueError("This rule cannot be triggered (disabled or active)")
        for relationship in self.checked_object_relationships.iterator():
            relationship.process_rule(evaluation_date, override_incident=override_incident)
        if not silent_notification:
            self.notify(evaluation_date)

    def get_permissions_for_user(self, user, created: datetime | None = None) -> dict[str, bool]:
        permissions = super().get_permissions_for_user(user, created)

        if user.has_perm(self.view_perm_str):
            permissions[self.view_perm_str] = False

        return permissions

    def get_or_create_incident(
        self,
        evaluation_date: date,
        incident_severity: "RiskIncidentType",
        breached_object: Any,
        breached_object_repr: str,
    ) -> tuple[RiskIncident, bool]:
        """
        Utility function to get or create incident base on the given breached object

        Args:
            rule:  The rule this incident happens
            evaluation_date: The incident date
            incident_severity: The incident severity
            breached_object: The breached object (i.e. the object that triggers the incident)
            breached_object_repr: Its string representation

        Returns:
            A tuple (incident, is_created)
        """

        # Consider that if an incident happens one business day in the future or past, it is continue
        date_range = DateRange(  # type: ignore
            (evaluation_date - BDay(1)).date(), (evaluation_date + BDay(1)).date(), "[]"
        )
        incidents = RiskIncident.objects.filter(rule=self, date_range__overlap=date_range)
        if breached_object:
            # If a breached_object is provided, we lookup over it
            incidents = incidents.filter(
                breached_content_type=ContentType.objects.get_for_model(breached_object),
                breached_object_id=breached_object.id,
            )
        else:
            # Otherwise, we lookup over its string representation. Might need to change as this is not very robust
            incidents = incidents.filter(
                breached_object_repr=breached_object_repr,
            )
        if incident := incidents.first():
            incident.date_range = DateRange(
                lower=incident.date_range.lower, upper=(evaluation_date + timedelta(days=1))
            )  # type: ignore
            if self.automatically_close_incident:
                incident.status = RiskIncident.Status.CLOSED
            incident.save()
            return incident, False
        else:
            return (
                RiskIncident.objects.create(
                    rule=self,
                    date_range=DateRange(lower=evaluation_date, upper=(evaluation_date + timedelta(days=1))),  # type: ignore
                    breached_content_type=ContentType.objects.get_for_model(breached_object)
                    if breached_object
                    else None,
                    breached_object_id=breached_object.id if breached_object else None,
                    breached_object_repr=breached_object_repr,
                    severity=incident_severity,
                    status=RiskIncident.Status.CLOSED
                    if self.automatically_close_incident
                    else RiskIncident.Status.OPEN,
                ),
                True,
            )

    def notify(self, evaluation_date: date):
        """
        Create the notification for this rule relationship and all implied persons
        """
        # Check if the incident needs to be notified
        incidents = RiskIncident.objects.filter(
            models.Q(checked_object_relationships__rule_check__evaluation_date=evaluation_date)
            & models.Q(rule=self)
            & (models.Q(status=RiskIncident.Status.OPEN) | models.Q(is_notified=False))
        )
        if not self.is_silent:
            for threshold in self.thresholds.filter(
                severity__is_informational=False
            ):  # ignore informational incident from being ignored
                threshold_incidents = incidents.filter(severity=threshold.severity)
                notified_users = threshold.get_notifiable_users()
                if notified_users.exists() and threshold_incidents.exists():
                    evaluation_date_sub_incidents = CheckedObjectIncidentRelationship.objects.filter(
                        incident__in=threshold_incidents,
                        rule_check__evaluation_date=evaluation_date,
                        severity=threshold.severity,
                    ).order_by("rule_check__checked_object_repr")
                    if evaluation_date_sub_incidents.exists():
                        breached_content_types = ContentType.objects.filter(
                            id__in=evaluation_date_sub_incidents.values("incident__breached_content_type")
                        )
                        breached_content_type_name = "Object"
                        if breached_content_types.count() == 1:
                            breached_content_type_name = breached_content_types.first().name
                        html = get_template("risk_management/incident_notification.html").render(
                            {
                                "evaluation_date_sub_incidents": evaluation_date_sub_incidents,
                                "rule": self,
                                "threshold": threshold,
                                "evaluation_date": evaluation_date,
                                "check_object_content_type_name": "Checked " + self.checked_object_representation,
                                "breached_content_type_name": breached_content_type_name,
                            }
                        )
                        for user in notified_users:
                            send_notification(
                                code="wbcompliance.riskincident.notify",
                                title=_("{} Broken Rule: {} as of {}").format(
                                    threshold.severity.name, self.name, evaluation_date.strftime("%d.%m.%Y")
                                ),
                                body=html,
                                user=user,
                                endpoint=reverse("wbcompliance:riskrule-detail", args=[self.id]),
                            )
        incidents.update(is_notified=True)

    @classmethod
    def get_rules_for_object(cls, obj) -> "models.QuerySet[RiskRule]":
        """
        Returns a Queryset of document linked to the passed object

        Args:
            obj: The related object

        Returns:
            A queryset of documents
        """
        rule_ids = RuleCheckedObjectRelationship.objects.filter(
            checked_object_content_type=ContentType.objects.get_for_model(obj), checked_object_id=obj.id
        ).values("rule")
        return cls.objects.filter(id__in=rule_ids)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcompliance:riskrule"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcompliance:riskrulerepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{name}}"


@receiver(pre_save, sender="wbcompliance.RuleThreshold")
def pre_save_risk_threshold(sender, instance, **kwargs):
    with suppress(RuleThreshold.DoesNotExist):
        pre_save_instance = RuleThreshold.objects.get(id=instance.id)
        RiskIncident.objects.filter(rule=instance.rule, severity=pre_save_instance.severity).update(
            severity=instance.severity
        )
        CheckedObjectIncidentRelationship.objects.filter(
            incident__rule=instance.rule, severity=pre_save_instance.severity
        ).update(severity=instance.severity)


@receiver(pre_delete, sender="wbcompliance.RuleThreshold")
def pre_delete_risk_threshold(sender, instance, **kwargs):
    RiskIncident.objects.filter(rule=instance.rule, severity=instance.severity).delete()
    CheckedObjectIncidentRelationship.objects.filter(incident__rule=instance.rule, severity=instance.severity).delete()


@shared_task
def process_rule_as_task(rule_id: int, evaluation_date: date, override_incident: bool | None = False):
    """
    Async task to process rule
    """
    rule = RiskRule.objects.get(id=rule_id)
    rule.process_rule(evaluation_date, override_incident=override_incident)



# Generated by Django 5.0.9 on 2024-11-27 15:14

from contextlib import suppress

import django.db.models.deletion
from django.db import migrations, models


def set_activation_date(*args):
    from wbcompliance.models.risk_management import RiskCheck, RiskRule

    for rule in RiskRule.objects.filter(activation_date__isnull=True):
        with suppress(RiskCheck.DoesNotExist):
            rule.activation_date = rule.checks.earliest("evaluation_date").evaluation_date
            rule.save()


class Migration(migrations.Migration):
    dependencies = [
        ("wbcompliance", "0018_alter_rulecheckedobjectrelationship_unique_together"),
    ]

    operations = [
        migrations.CreateModel(
            name="RuleGroup",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("key", models.CharField(max_length=255, unique=True)),
                ("name", models.CharField(max_length=255)),
            ],
        ),
        migrations.AddField(
            model_name="riskrule",
            name="activation_date",
            field=models.DateField(blank=True, null=True, verbose_name="Activation Date"),
        ),
        migrations.AddField(
            model_name="riskrule",
            name="frequency",
            field=models.CharField(
                blank=True,
                help_text="The Evaluation Frequency in RRULE format",
                max_length=56,
                null=True,
                verbose_name="Evaluation Frequency",
            ),
        ),
        migrations.AddField(
            model_name="rulebackend",
            name="rule_group",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="rules",
                to="wbcompliance.rulegroup",
            ),
        ),
        migrations.RunPython(set_activation_date),
    ]



import pandas as pd
from django.contrib import admin
from reversion_compare.admin import CompareVersionAdmin
from wbcore.admin import ExportCsvMixin

from wbcompliance.models import (
    ComplianceAction,
    ComplianceEvent,
    ComplianceTask,
    ComplianceTaskGroup,
    ComplianceTaskInstance,
    ComplianceType,
    ReviewComplianceTask,
)

from .utils import CustomImportCsvMixin


@admin.register(ComplianceTaskGroup)
class ComplianceTaskGroupAdmin(ExportCsvMixin, CustomImportCsvMixin, CompareVersionAdmin):
    list_display = ["name", "order"]

    def manipulate_df(self, df):
        df["name"] = df["name"].fillna("")
        return df

    def get_import_fields(self):
        return [
            "name",
        ]

    def process_model(self, model):
        if model.get("name"):
            _, created = self.model.objects.get_or_create(name=model.get("name"))
            return 1 if created else 0
        return 0


@admin.register(ComplianceTask)
class ComplianceTaskAdmin(ExportCsvMixin, CustomImportCsvMixin, CompareVersionAdmin):
    list_display = ["title", "occurrence", "active", "risk_level", "group", "type"]

    def manipulate_df(self, df):
        df["type"] = df["type_name"].apply(lambda x: ComplianceType.objects.filter(name__exact=x).first())
        df["group"] = df.apply(
            lambda x: ComplianceTaskGroup.objects.filter(name__exact=x["group_name"]).first(), axis=1
        )
        return df

    def get_import_fields(self):
        return ["type_name", "group_name", "occurrence", "title", "description"]

    def process_model(self, model):
        if model.get("title"):
            _, created = self.model.objects.update_or_create(
                title=model.get("title"),
                defaults={
                    "type": model.get("type"),
                    "group": model.get("group"),
                    "occurrence": model.get("occurrence"),
                    # 'description': model.get('description')
                },
            )
            return 1 if created else 0
        return 0


@admin.register(ComplianceTaskInstance)
class ComplianceTaskInstanceAdmin(ExportCsvMixin, CustomImportCsvMixin, CompareVersionAdmin):
    list_display = ["task", "occured", "status"]

    def manipulate_df(self, df):
        df["task"] = df.apply(lambda x: ComplianceTask.objects.filter(title__exact=x["task_name"]).first(), axis=1)
        df["occured"] = pd.to_datetime(df["occured"])
        return df

    def get_import_fields(self):
        return ["task_name", "occured", "status", "text"]

    def process_model(self, model):
        if (task := model.get("task")) and (occured := model.get("occured")):
            obj, created = self.model.objects.update_or_create(
                task=task,
                occured=occured,
                defaults={
                    "status": model.get("status"),
                    "text": model.get("text"),
                },
            )
            if created:
                obj.occured = occured
                obj.save()

            return 1 if created else 0
        return 0


@admin.register(ComplianceAction)
class ComplianceActionAdmin(CompareVersionAdmin):
    search_fields = ["title"]
    list_display = [
        "title",
        "deadline",
        "progress",
        "status",
        "type",
        "creator",
        "created",
        "changer",
        "last_modified",
    ]


@admin.register(ComplianceEvent)
class ComplianceEventAdmin(CompareVersionAdmin):
    search_fields = ["title"]
    list_display = ["title", "type", "level", "type", "confidential", "creator", "created", "changer", "last_modified"]


@admin.register(ReviewComplianceTask)
class ReviewComplianceTaskAdmin(CompareVersionAdmin):
    list_display = [
        "id",
        "year",
        "title",
        "from_date",
        "to_date",
        "status",
        "occurrence",
        "is_instance",
        "changer",
        "changed",
        "review_task",
        "occured",
    ]
    search_fields = ["title"]



from io import StringIO

import pandas as pd
from django.shortcuts import redirect, render
from reversion.errors import RevertError
from wbcore.admin import CsvImportForm, ImportCsvMixin


class CustomImportCsvMixin(ImportCsvMixin):
    def _import_csv(self, request, _sep=";"):
        if request.method == "POST":
            csv_file = request.FILES["csv_file"]

            str_text = ""
            for line in csv_file:
                str_text = str_text + line.decode()
            # Import csv as df
            df = pd.read_csv(StringIO(str_text), sep=_sep)
            # Sanitize dataframe
            df = df.where(pd.notnull(df), None)
            df = df.drop(df.columns.difference(self.get_import_fields()), axis=1)

            # Overide this function if there is foreign key ids in the dataframe
            df = self.manipulate_df(df)
            errors = 0
            revert_errors = 0
            nb_added = 0
            for model in df.to_dict("records"):
                # by default, process the modela as a create request. Can be override to change the behavior
                try:
                    nb_added += self.process_model(model)
                # https://django-reversion.readthedocs.io/en/stable/common-problems.html
                except RevertError:
                    revert_errors += 1
                except Exception as e:
                    print(e)  # noqa: T201
                    errors += 1
            msg = f"""Your csv file has been imported : {df.shape[0] - errors - revert_errors} imported
                 ({nb_added} added, {df.shape[0] - errors - revert_errors - nb_added} updated), {errors} errors,
                 {revert_errors} revert errors found due to failure to restore old versions
            """
            self.message_user(request, msg)
            return redirect("..")
        form = CsvImportForm()
        payload = {"form": form}
        return render(request, "wbcore/admin/csv_form.html", payload)



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig

from wbcompliance.models import ComplianceType, ReviewComplianceTask


class ComplianceTaskGroupEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        if ComplianceType.is_administrator(self.request.user):
            return super().get_endpoint()
        return None


class ComplianceTaskEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        if ComplianceType.is_administrator(self.request.user):
            return super().get_endpoint()
        return None


class ComplianceTaskComplianceTaskGroupEndpointConfig(ComplianceTaskEndpointConfig):
    def get_endpoint(self, **kwargs):
        if group_id := self.view.kwargs.get("group_id", None):
            return reverse(
                "wbcompliance:compliancetaskgroup-compliancetask-list", args=[group_id], request=self.request
            )
        return None


class ComplianceTaskInstanceEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        if ComplianceType.is_administrator(self.request.user):
            return super().get_endpoint()
        return None

    def get_create_endpoint(self, **kwargs):
        return None

    def get_delete_endpoint(self, **kwargs):
        if self.request.user.is_superuser:
            return super().get_delete_endpoint()
        return None


class ComplianceTaskInstanceComplianceTaskEndpointConfig(ComplianceTaskInstanceEndpointConfig):
    def get_endpoint(self, **kwargs):
        if ComplianceType.is_administrator(self.request.user):
            return reverse(
                "wbcompliance:compliancetask-compliancetaskinstance-list",
                args=[self.view.kwargs["task_id"]],
                request=self.request,
            )
        return None


class ComplianceTaskMatrixEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        if ComplianceType.is_administrator(self.request.user):
            return reverse("wbcompliance:compliancetaskmatrix-list", [], request=self.request)
        return None

    def get_create_endpoint(self, **kwargs):
        return None

    def get_instance_endpoint(self, **kwargs):
        return None


class ComplianceActionEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        if ComplianceType.is_administrator(self.request.user):
            return super().get_endpoint()
        return None


class ComplianceEventEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return super().get_endpoint()

    def get_instance_endpoint(self, **kwargs):
        if self.instance:
            obj = self.view.get_object()
            if not ComplianceType.is_administrator(self.request.user) and obj.creator != self.request.user.profile:
                return None
        return super().get_instance_endpoint()

    def get_delete_endpoint(self, **kwargs):
        if ComplianceType.is_administrator(self.request.user):
            return super().get_delete_endpoint()
        return None


class ReviewComplianceTaskEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        if ComplianceType.is_administrator(self.request.user):
            return super().get_endpoint()
        return None

    def get_instance_endpoint(self, **kwargs):
        if self.instance:
            if not ComplianceType.is_administrator(self.request.user):
                return None
            if obj_id := self.view.kwargs.get("pk", None):
                obj = ReviewComplianceTask.objects.get(id=obj_id)
                if obj.status != ReviewComplianceTask.Status.DRAFT:
                    return None
        return super().get_instance_endpoint()

    def get_delete_endpoint(self, **kwargs):
        if not ComplianceType.is_administrator(self.request.user):
            return None
        if obj_id := self.view.kwargs.get("pk", None):
            obj = ReviewComplianceTask.objects.get(id=obj_id)
            if obj.status != ReviewComplianceTask.Status.DRAFT or obj.is_instance:
                return None
        return super().get_delete_endpoint()


class AbstractComplianceTaskReviewEndpointConfig(EndpointViewConfig):
    def get_instance_endpoint(self, **kwargs):
        if self.instance:
            if not ComplianceType.is_administrator(self.request.user):
                return None
            if review_id := self.view.kwargs.get("review_id", None):
                obj = ReviewComplianceTask.objects.get(id=review_id)
                if obj.status != ReviewComplianceTask.Status.DRAFT:
                    return None
        return super().get_instance_endpoint()

    def get_create_endpoint(self, **kwargs):
        if not ComplianceType.is_administrator(self.request.user):
            return None
        if review_id := self.view.kwargs.get("review_id", None):
            obj = ReviewComplianceTask.objects.get(id=review_id)
            if obj.status != ReviewComplianceTask.Status.DRAFT:
                return None
        return super().get_create_endpoint()

    def get_delete_endpoint(self, **kwargs):
        if not ComplianceType.is_administrator(self.request.user):
            return None
        if review_id := self.view.kwargs.get("review_id", None):
            obj = ReviewComplianceTask.objects.get(id=review_id)
            if obj.status != ReviewComplianceTask.Status.DRAFT:
                return None
        return super().get_delete_endpoint()


class ComplianceTaskReviewNoGroupEndpointConfig(AbstractComplianceTaskReviewEndpointConfig):
    def get_endpoint(self, **kwargs):
        if ComplianceType.is_administrator(self.request.user):
            if review_id := self.view.kwargs.get("review_id", None):
                return reverse(
                    "wbcompliance:review-compliancetasknogroup-list", args=[review_id], request=self.request
                )
        return None


class ComplianceTaskReviewGroupEndpointConfig(AbstractComplianceTaskReviewEndpointConfig):
    def get_endpoint(self, **kwargs):
        if (
            ComplianceType.is_administrator(self.request.user)
            and (review_id := self.view.kwargs.get("review_id", None))
            and (group_id := self.view.kwargs.get("group_id", None))
        ):
            return reverse(
                "wbcompliance:review-compliancetaskgroup-list", args=[review_id, group_id], request=self.request
            )
        return None


class ComplianceTaskInstanceReviewNoGroupEndpointConfig(ComplianceTaskInstanceEndpointConfig):
    def get_endpoint(self, **kwargs):
        if ComplianceType.is_administrator(self.request.user) and (
            review_id := self.view.kwargs.get("review_id", None)
        ):
            return reverse(
                "wbcompliance:review-compliancetaskinstancenogroup-list", args=[review_id], request=self.request
            )
        return None


class ComplianceTaskInstanceReviewGroupEndpointConfig(ComplianceTaskInstanceEndpointConfig):
    def get_endpoint(self, **kwargs):
        if ComplianceType.is_administrator(self.request.user):
            if (review_id := self.view.kwargs.get("review_id", None)) and (
                group_id := self.view.kwargs.get("group_id", None)
            ):
                return reverse(
                    "wbcompliance:review-compliancetaskinstancegroup-list",
                    args=[review_id, group_id],
                    request=self.request,
                )
        return None



from django.contrib import admin
from reversion_compare.admin import CompareVersionAdmin

from wbcompliance.models import ComplianceType


@admin.register(ComplianceType)
class ComplianceTypeAdmin(CompareVersionAdmin):
    list_display = ["name"]



from .checks import RiskCheckEndpointConfig
from .incidents import (
    CheckedObjectIncidentRelationshipEndpointConfig,
    CheckedObjectIncidentRelationshipRiskRuleEndpointConfig,
    RiskIncidentEndpointConfig,
)
from .rules import (
    RiskRuleEndpointConfig,
    RuleCheckedObjectRelationshipRiskRuleEndpointConfig,
    RuleThresholdRiskRuleEndpointConfig,
)
from .tables import RiskManagementIncidentTableEndpointConfig



from django.db.models import Q
from django.utils.translation import gettext_lazy as _
from wbcore import filters as wb_filters

from wbcompliance.models.risk_management import (
    CheckedObjectIncidentRelationship,
    RiskIncident,
    RiskIncidentType,
    RiskRule,
)
from wbcompliance.models.risk_management.rules import RuleGroup


def get_default_severity(*args, **kwargs):
    non_informational_qs = RiskIncidentType.objects.exclude(is_informational=True)
    if non_informational_qs.exists():
        return list(non_informational_qs.values_list("id", flat=True))
    return []


class RiskIncidentFilterSet(wb_filters.FilterSet):
    date_range = wb_filters.DateRangeFilter(label="Date Range")
    status = wb_filters.ChoiceFilter(
        initial=RiskIncident.Status.OPEN, choices=RiskIncident.Status.choices, label=_("Label")
    )
    rule = wb_filters.ModelMultipleChoiceFilter(
        label=_("Rule"),
        queryset=RiskRule.objects.all(),
        endpoint=RiskRule.get_representation_endpoint(),
        value_key=RiskRule.get_representation_value_key(),
        label_key=RiskRule.get_representation_label_key(),
    )

    severity = wb_filters.ModelMultipleChoiceFilter(
        label=_("Severity"),
        queryset=RiskIncidentType.objects.all(),
        endpoint=RiskIncidentType.get_representation_endpoint(),
        value_key=RiskIncidentType.get_representation_value_key(),
        label_key=RiskIncidentType.get_representation_label_key(),
        initial=get_default_severity,
    )
    checked_object_relationships = wb_filters.MultipleChoiceContentTypeFilter(
        label="Checked Objects",
        field_name="checked_object_relationships",
        object_id_label="checked_object_relationships__rule_check__checked_object_id",
        content_type_label="checked_object_relationships__rule_check__checked_object_content_type",
        distinct=True,
        hidden=True,
    )

    object_repr = wb_filters.CharFilter(label="Object Representation", lookup_expr="icontains")

    rule_group = wb_filters.ModelChoiceFilter(
        label=_("Group"),
        queryset=RuleGroup.objects.all(),
        endpoint=RuleGroup.get_representation_endpoint(),
        value_key=RuleGroup.get_representation_value_key(),
        label_key=RuleGroup.get_representation_label_key(),
        method="filter_rule_group",
    )

    def filter_rule_group(self, queryset, label, value):
        if value:
            return queryset.filter(
                Q(rule__rule_backend__rule_group=value) | Q(rule__rule_backend__rule_group__isnull=True)
            )
        return queryset

    class Meta:
        model = RiskIncident
        fields = {
            "status": ["exact"],
            "resolved_by": ["exact"],
            "comment": ["icontains"],
            "severity": ["exact"],
        }


class CheckedObjectIncidentRelationshipFilterSet(wb_filters.FilterSet):
    checked_object = wb_filters.CharFilter(
        method="filter_checked_object", field_name="checked_object", lookup_expr="icontains"
    )

    def filter_checked_object(self, queryset, label, value):
        if value:
            return queryset.filter(rule_check__checked_object_repr__icontains=value)
        return queryset

    class Meta:
        model = CheckedObjectIncidentRelationship
        fields = {
            "severity": ["exact"],
            "report": ["icontains"],
            "comment": ["icontains"],
            "resolved_by": ["exact"],
            "incident": ["exact"],
            "rule_check": ["exact"],
            "status": ["exact"],
        }



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class RiskIncidentEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        if rule_id := self.view.kwargs.get("rule_id"):
            return reverse(
                "wbcompliance:riskrule-incident-list",
                args=[rule_id],
                request=self.request,
            )
        return super().get_endpoint(**kwargs)

    def get_create_endpoint(self, **kwargs):
        return None

    def get_delete_endpoint(self, **kwargs):
        return None


class CheckedObjectIncidentRelationshipEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_list_endpoint(self, **kwargs):
        return reverse("wbcompliance:checkedobjectincidentrelationship-list", request=self.request)


class CheckedObjectIncidentRelationshipRiskRuleEndpointConfig(CheckedObjectIncidentRelationshipEndpointConfig):
    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbcompliance:riskincident-relationship-list",
            args=[self.view.kwargs["incident_id"]],
            request=self.request,
        )



# Generated by Django 5.0.8 on 2024-08-28 11:32

import django.core.serializers.json
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbcompliance", "0015_auto_20240103_0957"),
    ]

    operations = [
        migrations.AddField(
            model_name="checkedobjectincidentrelationship",
            name="report_details",
            field=models.JSONField(default=dict, blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder),
        ),
        migrations.AddField(
            model_name="rulebackend",
            name="incident_report_template",
            field=models.TextField(
                default="<ul>\n    {% for key, value in report_details.items %}\n        <li><strong>{{ key }}: </strong> {{ value }}</li>\n    {% endfor %}\n</ul>\n",
                verbose_name="Incident Template",
            ),
        ),
        migrations.AddField(
            model_name="checkedobjectincidentrelationship",
            name="breached_value",
            field=models.CharField(
                blank=True,
                help_text="The value that breached the rule threshold, can be None",
                max_length=128,
                null=True,
                verbose_name="Breached Value",
            ),
        ),
    ]



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class RiskCheckEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbcompliance:riskcheck-list",
            request=self.request,
        )



from django.utils.translation import gettext_lazy
from wbcore import serializers as wb_serializers
from wbcore.contrib.authentication.serializers import GroupRepresentationSerializer
from wbcore.contrib.directory.serializers import PersonRepresentationSerializer

from wbcompliance.models import ComplianceType


class ComplianceTypeRepresentationSerializer(wb_serializers.RepresentationSerializer):
    class Meta:
        model = ComplianceType
        fields = (
            "id",
            "name",
        )


class ComplianceTypeModelSerializer(wb_serializers.ModelSerializer):
    _in_charge = GroupRepresentationSerializer(source="in_charge", many=True)
    administrators = wb_serializers.PrimaryKeyRelatedField(
        label=gettext_lazy("Administrator"), many=True, read_only=True
    )
    _administrators = PersonRepresentationSerializer(many=True, source="administrators")

    class Meta:
        model = ComplianceType
        fields = ("id", "name", "description", "in_charge", "_in_charge", "administrators", "_administrators")



from typing import Optional

from django.utils.translation import gettext as _
from wbcore.enums import Unit
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

from wbcompliance.models.risk_management.checks import RiskCheck


class RiskCheckDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(
                    key="status_icon",
                    label=_("Status"),
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"color": RiskCheck.CheckStatus[value].color},
                            condition=("==", RiskCheck.CheckStatus[value].icon),
                        )
                        for value in RiskCheck.CheckStatus.values
                    ],
                    width=Unit.PIXEL(75),
                ),
                dp.Field(key="rule", label=_("Rule"), width=Unit.PIXEL(300)),
                dp.Field(key="evaluation_date", label=_("Evaluation Date"), width=Unit.PIXEL(150)),
                dp.Field(key="creation_datetime", label=_("Creation Time"), width=Unit.PIXEL(150)),
                dp.Field(key="incident_reports", label=_("Incidents"), width=Unit.PIXEL(400)),
            ],
            legends=[
                dp.Legend(
                    items=[
                        dp.LegendItem(icon=RiskCheck.CheckStatus[value].icon, label=RiskCheck.CheckStatus[value].label)
                        for value in RiskCheck.CheckStatus.values
                    ]
                )
            ],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display([[repeat_field(3, "status")], ["rule", "evaluation_date", "creation_datetime"]])



from django.contrib.contenttypes.models import ContentType
from django.db.models import Count, Q
from django.dispatch import receiver
from django.utils.translation import gettext_lazy as _
from wbcore import filters as wb_filters
from wbcore.signals.filters import add_filters

from wbcompliance.models.risk_management import CheckedObjectIncidentRelationship


def get_content_type_default(*args, **kwargs):
    content_type_id = (
        CheckedObjectIncidentRelationship.objects.values("rule_check__checked_object_content_type")
        .annotate(c=Count("rule_check__checked_object_content_type"))
        .order_by("c")
        .first()["rule_check__checked_object_content_type"]
    )
    return ContentType.objects.get(id=content_type_id).id


def filter_date_range(queryset, label, value):
    if value:
        return queryset.filter(
            Q(rule_check__evaluation_date__gte=value.lower) & Q(rule_check__evaluation_date__lte=value.upper)
        )
    return queryset


class RiskManagementIncidentFilter(wb_filters.FilterSet):
    incident_date_range = wb_filters.DateRangeFilter(label=_("Date Range"), method=filter_date_range)
    only_open_incident = wb_filters.BooleanFilter(
        initial=True, method="filter_only_open_incident", label=_("Only open incidents")
    )

    checked_object_content_type = wb_filters.ModelChoiceFilter(
        required=True,
        method="filter_checked_object_content_type",
        queryset=ContentType.objects.all(),
        endpoint="wbcore:contenttyperepresentation-list",
        value_key="id",
        label_key="{{app_label}} | {{model}}",
        label=_("Checked Object Type"),
        initial=get_content_type_default,
        filter_params={"related_name_isnull": "risk_management_checked_objects"},
        clearable=False,
    )

    def filter_checked_object_content_type(self, queryset, label, value):
        if value:
            return queryset.filter(rule_check__checked_object_content_type=value)
        return queryset

    def filter_only_open_incident(self, queryset, name, value):
        if value:
            return queryset.filter(status=CheckedObjectIncidentRelationship.Status.OPEN)
        return queryset

    class Meta:
        model = CheckedObjectIncidentRelationship
        fields = {
            # "rule_backend": ["exact"],
            # "is_enable": ["exact"],
            # "only_passive_check_allowed": ["exact"],
            # "is_silent": ["exact"],
            # "is_mandatory": ["exact"],
        }


@receiver(add_filters, sender=RiskManagementIncidentFilter)
def add_checked_object_repr_filter(sender, request=None, *args, **kwargs):
    content_type = None
    if request and (content_type_id := request.GET.get("checked_object_content_type", None)):
        content_type = ContentType.objects.get(id=content_type_id)
    elif default_callback := getattr(sender, "default", None):
        content_type = ContentType.objects.get(id=default_callback())
    if content_type:

        def filter_checked_object_repr(queryset, label, value):
            if value:
                return queryset.filter(rule_check__checked_object_id=value.id)
            return queryset

        model_class = content_type.model_class()
        return {
            "checked_object_repr": wb_filters.ModelChoiceFilter(
                label=content_type.name,
                field_name="checked_object_repr",
                queryset=model_class.objects.all(),
                endpoint=model_class.get_representation_endpoint(),
                value_key=model_class.get_representation_value_key(),
                label_key=model_class.get_representation_label_key(),
                method=filter_checked_object_repr,
            )
        }
    else:

        def filter_checked_object_repr(queryset, label, value):
            if value:
                return queryset.filter(rule_check__checked_object_repr__icontains=value)
            return queryset

        return {
            "checked_object_repr": wb_filters.CharFilter(
                label=_("Checked Object Repr."),
                method=filter_checked_object_repr,
                lookup_expr="icontains",
                field_name="checked_object_repr",
            )
        }



# Generated by Django 4.1.7 on 2023-04-19 05:36

import django.contrib.postgres.fields.ranges
import django.core.validators
import django.db.models.deletion
import django.utils.timezone
import django_fsm
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("directory", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("auth", "0012_alter_user_first_name_max_length"),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="ComplianceForm",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("created", models.DateTimeField(auto_now_add=True, verbose_name="Created")),
                ("changed", models.DateTimeField(blank=True, null=True)),
                ("title", models.CharField(max_length=255, verbose_name="Title")),
                ("policy", models.TextField(blank=True, default="", null=True)),
                (
                    "status",
                    django_fsm.FSMField(
                        choices=[("DRAFT", "Draft"), ("PENDING", "Pending"), ("ACTIVE", "Active")],
                        default="DRAFT",
                        help_text="The Compliance Form status (default to Draft)",
                        max_length=50,
                        verbose_name="Status",
                    ),
                ),
                ("version", models.IntegerField(default=0)),
                (
                    "only_send_to_internals",
                    models.BooleanField(
                        default=True,
                        help_text="Sent the Compliance Form only to internal employees of the departments",
                    ),
                ),
                ("end", models.DateField(blank=True, null=True, verbose_name="End")),
                ("start", models.DateField(verbose_name="Start")),
                (
                    "assigned_to",
                    models.ManyToManyField(
                        blank=True,
                        help_text="List of group to whom this Compliance Form applies.            It is applied to all users if not specified",
                        related_name="compliance_forms",
                        to="auth.group",
                        verbose_name="Group to which the Compliance Form applies",
                    ),
                ),
            ],
            options={
                "verbose_name": "Compliance Form",
                "verbose_name_plural": "Compliance Forms",
                "permissions": [("admin_complianceform", "Can administer Compliance Forms")],
            },
        ),
        migrations.CreateModel(
            name="ComplianceFormSignature",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("version", models.IntegerField(default=0)),
                ("policy_text", models.TextField(blank=True, default="", null=True)),
                ("signed", models.DateTimeField(blank=True, null=True)),
                ("personal_notes", models.TextField(blank=True, default="", null=True)),
                (
                    "compliance_form",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="complianceforms",
                        to="wbcompliance.complianceform",
                    ),
                ),
                ("end", models.DateField(blank=True, null=True, verbose_name="End")),
                ("start", models.DateField(verbose_name="Start")),
            ],
            options={
                "verbose_name": "Compliance Form Signature",
                "verbose_name_plural": "Compliance Signatures",
            },
        ),
        migrations.CreateModel(
            name="ComplianceFormSection",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=255, verbose_name="Name section")),
                (
                    "compliance_form",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="compliance_forms",
                        to="wbcompliance.complianceform",
                        verbose_name="Compliance Form",
                    ),
                ),
            ],
            options={
                "verbose_name": "Section of the Compliance Form",
                "verbose_name_plural": "Sections of the Compliance Form",
            },
        ),
        migrations.CreateModel(
            name="ComplianceFormRule",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("text", models.TextField(default="")),
                ("ticked", models.BooleanField(default=False, verbose_name="Is applicable")),
                (
                    "section",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="rules",
                        to="wbcompliance.complianceformsection",
                        verbose_name="Section",
                    ),
                ),
            ],
            options={
                "verbose_name": "Rule of the section of the Compliance Form",
                "verbose_name_plural": "Rules of the section of the Compliance Form",
            },
        ),
        migrations.CreateModel(
            name="ComplianceFormSignatureSection",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=255, verbose_name="Name section")),
                (
                    "compliance_form_signature",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="compliance_form_signatures",
                        to="wbcompliance.complianceformsignature",
                        verbose_name="Compliance Form Signature",
                    ),
                ),
            ],
            options={
                "verbose_name": "Section of the Compliance Form Signature",
                "verbose_name_plural": "Sections of the Compliance Form Signature",
            },
        ),
        migrations.CreateModel(
            name="ComplianceFormSignatureRule",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("text", models.TextField(default="")),
                ("ticked", models.BooleanField(default=False, verbose_name="Is applicable")),
                ("comments", models.TextField(default="")),
                (
                    "section",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="rules",
                        to="wbcompliance.complianceformsignaturesection",
                        verbose_name="Section",
                    ),
                ),
            ],
            options={
                "verbose_name": "Rule of the section of the Compliance Form Signature",
                "verbose_name_plural": "Rules of the section of the Compliance Form Signature",
            },
        ),
        migrations.CreateModel(
            name="ComplianceFormType",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=255, verbose_name="Name")),
                (
                    "type",
                    models.CharField(
                        choices=[("TEXT", "Text"), ("FORM", "Form")],
                        default="TEXT",
                        max_length=32,
                        verbose_name="Type",
                    ),
                ),
            ],
            options={
                "verbose_name": "Compliance Form Type",
                "verbose_name_plural": "Compliance Form Types",
            },
        ),
        migrations.AddField(
            model_name="complianceform",
            name="type",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="wbcompliance.complianceformtype",
                verbose_name="Type",
            ),
        ),
        migrations.AlterField(
            model_name="complianceform",
            name="status",
            field=django_fsm.FSMField(
                choices=[("DRAFT", "Draft"), ("ACTIVATION_REQUESTED", "Activation Requested"), ("ACTIVE", "Active")],
                default="DRAFT",
                help_text="The Compliance Form status (default to Draft)",
                max_length=50,
                verbose_name="Status",
            ),
        ),
        migrations.CreateModel(
            name="ComplianceTaskGroup",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=255, verbose_name="Name")),
                ("order", models.PositiveIntegerField(blank=True, null=True, verbose_name="Order")),
            ],
            options={
                "verbose_name": "Compliance Task Group",
                "verbose_name_plural": "Compliance Task Groups",
            },
        ),
        migrations.CreateModel(
            name="ComplianceType",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=255, verbose_name="Name")),
                ("description", models.TextField(blank=True, default="", verbose_name="Description")),
            ],
            options={
                "verbose_name": "Compliance Type",
                "verbose_name_plural": "Compliance Types",
            },
        ),
        migrations.CreateModel(
            name="ComplianceAction",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("title", models.CharField(max_length=255, verbose_name="Title")),
                ("description", models.TextField(blank=True, default="", verbose_name="Description")),
                ("deadline", models.DateField(blank=True, null=True)),
                (
                    "progress",
                    models.FloatField(
                        default=0,
                        validators=[
                            django.core.validators.MinValueValidator(0.0),
                            django.core.validators.MaxValueValidator(1.0),
                        ],
                        verbose_name="Progress",
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("TO_BE_DONE", "To be done"),
                            ("WORK_IN_PROGRESS", "Work in Progress"),
                            ("DONE", "Done"),
                        ],
                        default="TO_BE_DONE",
                        max_length=32,
                        verbose_name="Status",
                    ),
                ),
                (
                    "type",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="wbcompliance.compliancetype",
                        verbose_name="Type",
                    ),
                ),
                ("active", models.BooleanField(default=True)),
                ("created", models.DateTimeField(auto_now_add=True, verbose_name="Created")),
                ("last_modified", models.DateTimeField(auto_now=True, verbose_name="Last modified")),
                (
                    "summary_description",
                    models.TextField(
                        blank=True,
                        help_text="Description for the Board of Directors",
                        null=True,
                        verbose_name="Summary Description",
                    ),
                ),
            ],
            options={
                "verbose_name": "Compliance Action",
                "verbose_name_plural": "Compliance Actions",
            },
        ),
        migrations.CreateModel(
            name="ComplianceEvent",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "type_event",
                    models.CharField(
                        choices=[("INCIDENT", "Incident"), ("INFO", "Info")],
                        default="INCIDENT",
                        max_length=32,
                        verbose_name="Type Event",
                    ),
                ),
                (
                    "level",
                    models.CharField(
                        choices=[("LOW", "Low"), ("MEDIUM", "Medium"), ("HIGH", "High")],
                        default="LOW",
                        max_length=32,
                        verbose_name="Level",
                    ),
                ),
                ("title", models.CharField(max_length=255, verbose_name="Title")),
                ("exec_summary", models.TextField(blank=True, default="", verbose_name="Executive Summary")),
                ("description", models.TextField(blank=True, default="", verbose_name="Description")),
                ("actions_taken", models.TextField(blank=True, default="", verbose_name="Actions Taken")),
                ("consequences", models.TextField(blank=True, default="", verbose_name="Consequences")),
                ("future_suggestions", models.TextField(blank=True, default="", verbose_name="Future Suggestions")),
                (
                    "type",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="wbcompliance.compliancetype",
                        verbose_name="Type",
                    ),
                ),
                ("active", models.BooleanField(default=True)),
                ("created", models.DateTimeField(auto_now_add=True, verbose_name="Created")),
                ("confidential", models.BooleanField(default=False)),
                ("last_modified", models.DateTimeField(auto_now=True, verbose_name="Last modified")),
                (
                    "exec_summary_board",
                    models.TextField(blank=True, null=True, verbose_name="Executive Summary for the Board"),
                ),
            ],
            options={
                "verbose_name": "Compliance Event",
                "verbose_name_plural": "Compliance Events",
            },
        ),
        migrations.CreateModel(
            name="ReviewComplianceTask",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("title", models.CharField(max_length=255, verbose_name="Title")),
                ("from_date", models.DateField(blank=True, null=True, verbose_name="From")),
                ("to_date", models.DateField(blank=True, null=True, verbose_name="To")),
                ("description", models.TextField(blank=True, default="", verbose_name="Description")),
                (
                    "year",
                    models.IntegerField(
                        blank=True,
                        null=True,
                        validators=[
                            django.core.validators.MinValueValidator(1000),
                            django.core.validators.MaxValueValidator(9999),
                        ],
                        verbose_name="Year",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, verbose_name="Created")),
                ("changed", models.DateTimeField(auto_now=True, null=True)),
                (
                    "status",
                    django_fsm.FSMField(
                        choices=[
                            ("DRAFT", "Draft"),
                            ("VALIDATION_REQUESTED", "Validation Requested"),
                            ("VALIDATED", "Validated"),
                        ],
                        default="DRAFT",
                        max_length=50,
                        verbose_name="Status",
                    ),
                ),
                (
                    "occurance",
                    models.CharField(
                        choices=[
                            ("YEARLY", "Yearly"),
                            ("QUARTERLY", "Quarterly"),
                            ("MONTHLY", "Monthly"),
                            ("NEVER", "Never"),
                        ],
                        default="MONTHLY",
                        max_length=32,
                        verbose_name="Occurance",
                    ),
                ),
                ("computed_str", models.CharField(blank=True, max_length=512, null=True)),
                ("is_instance", models.BooleanField(default=False)),
                (
                    "review_task",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="review_tasks",
                        to="wbcompliance.reviewcompliancetask",
                        verbose_name="Linked Indicator Report",
                    ),
                ),
                ("occured", models.DateTimeField(blank=True, null=True, verbose_name="Occured Instance Report")),
            ],
            options={
                "verbose_name": "Review Compliance Task",
                "verbose_name_plural": "Review Compliance Tasks",
            },
        ),
        migrations.CreateModel(
            name="ComplianceTask",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("title", models.CharField(max_length=255, verbose_name="Title")),
                ("description", models.TextField(blank=True, default="", verbose_name="Description")),
                (
                    "occurance",
                    models.CharField(
                        choices=[
                            ("YEARLY", "Yearly"),
                            ("QUARTERLY", "Quarterly"),
                            ("MONTHLY", "Monthly"),
                            ("NEVER", "Never"),
                        ],
                        default="MONTHLY",
                        max_length=32,
                        verbose_name="Occurance",
                    ),
                ),
                ("active", models.BooleanField(default=True)),
                (
                    "group",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="tasks_related",
                        to="wbcompliance.compliancetaskgroup",
                        verbose_name="Group",
                    ),
                ),
                (
                    "type",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="wbcompliance.compliancetype",
                        verbose_name="Type",
                    ),
                ),
                (
                    "review",
                    models.ManyToManyField(
                        blank=True,
                        help_text="list of reports that contain this task",
                        related_name="tasks",
                        to="wbcompliance.reviewcompliancetask",
                        verbose_name="Indicator Reports",
                    ),
                ),
                ("remarks", models.TextField(blank=True, null=True, verbose_name="Remarks")),
                (
                    "risk_level",
                    models.CharField(
                        blank=True,
                        choices=[("LOW", "Low"), ("MEDIUM", "Medium"), ("HIGH", "High")],
                        max_length=32,
                        null=True,
                        verbose_name="Risk Level",
                    ),
                ),
            ],
            options={
                "verbose_name": "Compliance Task",
                "verbose_name_plural": "Compliance Tasks",
            },
        ),
        migrations.CreateModel(
            name="ComplianceTaskInstance",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("occured", models.DateField(auto_now_add=True)),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("NOT_CHECKED", "Not Checked"),
                            ("WARNING", "Warning"),
                            ("FOR_INFO", "For Info"),
                            ("NOTHING_TO_REPORT", "Nothing to Report"),
                            ("BREACH", "Breach"),
                        ],
                        default="NOT_CHECKED",
                        max_length=32,
                        verbose_name="Status",
                    ),
                ),
                ("text", models.TextField(blank=True, default="", verbose_name="Text")),
                (
                    "task",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="task_instances_related",
                        to="wbcompliance.compliancetask",
                        verbose_name="Compliance Task",
                    ),
                ),
                (
                    "summary_text",
                    models.TextField(
                        blank=True,
                        help_text="Information for the Board of Directors",
                        null=True,
                        verbose_name="Summary Text",
                    ),
                ),
                (
                    "review",
                    models.ManyToManyField(
                        blank=True,
                        help_text="list of indicator instance reports that contain this task",
                        related_name="task_instances",
                        to="wbcompliance.reviewcompliancetask",
                        verbose_name="Indicator Instance Reports",
                    ),
                ),
            ],
            options={
                "verbose_name": "Compliance Task Instance",
                "verbose_name_plural": "Compliance Task Instances",
            },
        ),
        migrations.RemoveField(
            model_name="complianceform",
            name="assigned_to",
        ),
        migrations.AddField(
            model_name="complianceform",
            name="assigned_to",
            field=models.ManyToManyField(
                blank=True,
                help_text="List of group to whom this Compliance Form applies.            It is applied to all users if not specified",
                related_name="compliance_forms",
                to="auth.group",
                verbose_name="Group to which the Compliance Form applies",
            ),
        ),
        migrations.CreateModel(
            name="RiskCheck",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("computed_str", models.CharField(blank=True, max_length=512, null=True)),
                (
                    "creation_datetime",
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text="Time at which the check was created/triggered",
                        verbose_name="Creation Date",
                    ),
                ),
                (
                    "evaluation_date",
                    models.DateField(
                        help_text="The date at which the rule was evaluated", verbose_name="Evaluation Date"
                    ),
                ),
            ],
            options={
                "verbose_name": "Risk Check",
                "verbose_name_plural": "Risk Checks",
            },
        ),
        migrations.CreateModel(
            name="RiskIncident",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("computed_str", models.CharField(blank=True, max_length=512, null=True)),
                (
                    "severity",
                    models.CharField(
                        choices=[("LOW", "Low"), ("MEDIUM", "Medium"), ("HIGH", "High")],
                        default="LOW",
                        help_text="The incident Severity",
                        max_length=8,
                        verbose_name="Severity",
                    ),
                ),
                ("comment", models.TextField(blank=True, null=True, verbose_name="Comment")),
                (
                    "date_range",
                    django.contrib.postgres.fields.ranges.DateRangeField(
                        blank=True, help_text="The incident spans date interval", null=True
                    ),
                ),
                ("breached_object_id", models.PositiveIntegerField(blank=True, null=True)),
                (
                    "breached_object_repr",
                    models.CharField(
                        blank=True,
                        help_text="String Representation of the breached object",
                        max_length=128,
                        null=True,
                        verbose_name="Breached Object Representation",
                    ),
                ),
                (
                    "status",
                    django_fsm.FSMField(
                        choices=[
                            ("OPEN", "Open"),
                            ("RESOLVED", "Resolved"),
                            ("IGNORED", "Ignored"),
                            ("CLOSED", "Closed"),
                        ],
                        default="OPEN",
                        max_length=50,
                        verbose_name="Status",
                    ),
                ),
                (
                    "breached_content_type",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "verbose_name": "Risk Incident",
                "verbose_name_plural": "Risk Incidents",
            },
        ),
        migrations.CreateModel(
            name="RiskRule",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "permission_type",
                    models.CharField(
                        choices=[("INTERNAL", "Internal"), ("PUBLIC", "Public"), ("PRIVATE", "Private")],
                        default="PRIVATE",
                        max_length=8,
                    ),
                ),
                ("name", models.CharField(max_length=512)),
                ("description", models.CharField(default="", max_length=516, verbose_name="Quick Description")),
                ("is_enable", models.BooleanField(default=True, verbose_name="Enabled")),
                (
                    "is_passive",
                    models.BooleanField(
                        default=True,
                        help_text="If False, This rule can only be triggered through signal",
                        verbose_name="Passive",
                    ),
                ),
                (
                    "is_silent",
                    models.BooleanField(
                        default=True, help_text="If true, the notification won't be send through System nor Mail"
                    ),
                ),
                (
                    "is_mandatory",
                    models.BooleanField(
                        default=False,
                        help_text="A mandatory rule cannot be modified by anyone other than the administrators",
                    ),
                ),
                ("parameters", models.JSONField(blank=True, default=dict)),
                (
                    "creator",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="created_%(app_label)s_%(class)s",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "Risk Rule",
                "verbose_name_plural": "Risk Rules",
                "abstract": False,
                "default_permissions": ("add", "change", "delete", "view", "administrate"),
            },
        ),
        migrations.AlterField(
            model_name="complianceform",
            name="only_send_to_internals",
            field=models.BooleanField(default=True, help_text="Sent the Compliance Form only to internal users"),
        ),
        migrations.CreateModel(
            name="TriggererIncidentRelationship",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("computed_str", models.CharField(blank=True, max_length=512, null=True)),
                (
                    "severity",
                    models.CharField(
                        choices=[("LOW", "Low"), ("MEDIUM", "Medium"), ("HIGH", "High")],
                        default="LOW",
                        help_text="The incident Severity",
                        max_length=8,
                        verbose_name="Severity",
                    ),
                ),
                ("comment", models.TextField(blank=True, null=True, verbose_name="Comment")),
                ("report", models.TextField(blank=True, null=True, verbose_name="Report")),
                (
                    "status",
                    django_fsm.FSMField(
                        choices=[
                            ("OPEN", "Open"),
                            ("RESOLVED", "Resolved"),
                            ("IGNORED", "Ignored"),
                            ("CLOSED", "Closed"),
                        ],
                        default="OPEN",
                        max_length=50,
                        verbose_name="Status",
                    ),
                ),
                (
                    "incident",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="triggerer_relationships",
                        to="wbcompliance.riskincident",
                    ),
                ),
                (
                    "rule_check",
                    models.ForeignKey(
                        help_text="The check that opened this incident",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="incidents",
                        to="wbcompliance.riskcheck",
                        verbose_name="Check",
                    ),
                ),
            ],
            options={
                "verbose_name": "Incident to Triggerer relationship",
                "verbose_name_plural": "Incident to Triggerer relationships",
            },
        ),
        migrations.CreateModel(
            name="RuleTriggererRelationship",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("computed_str", models.CharField(blank=True, max_length=512, null=True)),
                ("triggerer_id", models.PositiveIntegerField()),
                ("triggerer_repr", models.CharField(blank=True, max_length=256, null=True)),
                (
                    "rule",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="triggerer_relationships",
                        to="wbcompliance.riskrule",
                    ),
                ),
                (
                    "triggerer_content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="risk_management_triggerers",
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "verbose_name": "Triggerer to Rule relationship",
                "verbose_name_plural": "Triggerer to Rule relationships",
            },
        ),
        migrations.CreateModel(
            name="RuleThreshold",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("computed_str", models.CharField(blank=True, max_length=512, null=True)),
                (
                    "range",
                    django.contrib.postgres.fields.ranges.DecimalRangeField(
                        help_text="The range which triggers the specified severity. null bound represent infinity",
                        verbose_name="Threshold range",
                    ),
                ),
                (
                    "severity",
                    models.CharField(
                        choices=[("LOW", "Low"), ("MEDIUM", "Medium"), ("HIGH", "High")],
                        default="LOW",
                        help_text="The Triggered Severity when the rule is within the threshold range",
                        max_length=8,
                        verbose_name="Triggered Severity",
                    ),
                ),
                (
                    "rule",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="thresholds",
                        to="wbcompliance.riskrule",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="RuleBackend",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=128)),
                ("backend_class_path", models.CharField(max_length=512)),
                ("backend_class_name", models.CharField(default="RuleBackend", max_length=128)),
                (
                    "allowed_triggerer_content_type",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "verbose_name": "Rule Backend",
                "verbose_name_plural": "Rule Backends",
            },
        ),
        migrations.AddField(
            model_name="riskrule",
            name="rule_backend",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, related_name="rules", to="wbcompliance.rulebackend"
            ),
        ),
        migrations.AddField(
            model_name="riskincident",
            name="rule",
            field=models.ForeignKey(
                help_text="The rule that opened this incident",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="incidents",
                to="wbcompliance.riskrule",
                verbose_name="Rule",
            ),
        ),
        migrations.AddField(
            model_name="riskcheck",
            name="rule_triggerer_relationship",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="checks",
                to="wbcompliance.ruletriggererrelationship",
                verbose_name="Rule-Triggerer Relationship",
            ),
        ),
        migrations.AddIndex(
            model_name="triggererincidentrelationship",
            index=models.Index(fields=["incident", "rule_check"], name="wbcomplianc_inciden_e01fa0_idx"),
        ),
        migrations.AlterUniqueTogether(
            name="triggererincidentrelationship",
            unique_together={("incident", "rule_check")},
        ),
        migrations.AddIndex(
            model_name="ruletriggererrelationship",
            index=models.Index(
                fields=["rule", "triggerer_content_type", "triggerer_id"], name="wbcomplianc_rule_id_c5fc0f_idx"
            ),
        ),
        migrations.AlterUniqueTogether(
            name="rulethreshold",
            unique_together={("rule", "severity")},
        ),
        migrations.AlterUniqueTogether(
            name="riskrule",
            unique_together={("rule_backend", "parameters")},
        ),
        migrations.AddField(
            model_name="rulethreshold",
            name="notifiable_groups",
            field=models.ManyToManyField(
                blank=True,
                help_text="Notified Groups for this rule and severity",
                related_name="notified_rule_thresholds",
                to="auth.group",
                verbose_name="Notifiable Groups",
            ),
        ),
        migrations.AddField(
            model_name="riskrule",
            name="apply_to_all_active_relationships",
            field=models.BooleanField(
                default=False,
                help_text="If True, will keep this rule syncrhonize with the backend definition of all active relationship",
            ),
        ),
        migrations.AddField(
            model_name="riskrule",
            name="automatically_close_incident",
            field=models.BooleanField(
                default=False, help_text="If True, this rule will automatically close all encountered incidents"
            ),
        ),
        migrations.AlterField(
            model_name="reviewcompliancetask",
            name="computed_str",
            field=models.CharField(blank=True, max_length=512, null=True, verbose_name="Name"),
        ),
        migrations.AlterField(
            model_name="riskcheck",
            name="computed_str",
            field=models.CharField(blank=True, max_length=512, null=True, verbose_name="Name"),
        ),
        migrations.AlterField(
            model_name="riskincident",
            name="computed_str",
            field=models.CharField(blank=True, max_length=512, null=True, verbose_name="Name"),
        ),
        migrations.AlterField(
            model_name="rulethreshold",
            name="computed_str",
            field=models.CharField(blank=True, max_length=512, null=True, verbose_name="Name"),
        ),
        migrations.AlterField(
            model_name="ruletriggererrelationship",
            name="computed_str",
            field=models.CharField(blank=True, max_length=512, null=True, verbose_name="Name"),
        ),
        migrations.AlterField(
            model_name="triggererincidentrelationship",
            name="computed_str",
            field=models.CharField(blank=True, max_length=512, null=True, verbose_name="Name"),
        ),
        migrations.CreateModel(
            name="RiskIncidentType",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=128)),
                ("severity_order", models.PositiveIntegerField(default=0, unique=True)),
                ("color", models.CharField(default="rgb(247, 232, 113)", max_length=20, verbose_name="Color")),
                ("is_ignorable", models.BooleanField(default=True, verbose_name="Can be ignored")),
                ("is_automatically_closed", models.BooleanField(default=False, verbose_name="Automatically closed")),
                (
                    "is_informational",
                    models.BooleanField(
                        default=False,
                        help_text="If true, the associated rule is not considered an incident",
                        verbose_name="Only Informational",
                    ),
                ),
            ],
            options={
                "verbose_name": "Risk Incident Type",
                "verbose_name_plural": "Risk Incidents Type",
                "ordering": ("severity_order",),
            },
        ),
        migrations.AddField(
            model_name="riskincident",
            name="is_notified",
            field=models.BooleanField(
                default=False,
                help_text="True if the incident is already notified to the users",
                verbose_name="Notified",
            ),
        ),
        migrations.AddField(
            model_name="rulethreshold",
            name="upgradable_after_days",
            field=models.PositiveIntegerField(
                blank=True,
                help_text="If set to a positive integer, the resulting incident will be elevated to the next rule severity after an incident remains open this number of days",
                null=True,
                verbose_name="Upgradable to next severity after X Days",
            ),
        ),
        migrations.RenameField(
            model_name="riskincident",
            old_name="severity",
            new_name="old_severity",
        ),
        migrations.RenameField(
            model_name="rulethreshold",
            old_name="severity",
            new_name="old_severity",
        ),
        migrations.RenameField(
            model_name="triggererincidentrelationship",
            old_name="severity",
            new_name="old_severity",
        ),
        migrations.AddField(
            model_name="rulethreshold",
            name="severity",
            field=models.ForeignKey(
                blank=True,
                help_text="The Triggered Severity when the rule is within the threshold range",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="thresholds",
                to="wbcompliance.riskincidenttype",
                verbose_name="Triggered Severity",
            ),
        ),
        migrations.RemoveField(
            model_name="riskincident",
            name="old_severity",
        ),
        migrations.RemoveField(
            model_name="triggererincidentrelationship",
            name="old_severity",
        ),
        migrations.AlterUniqueTogether(
            name="rulethreshold",
            unique_together={("rule", "severity")},
        ),
        migrations.RemoveField(
            model_name="rulethreshold",
            name="old_severity",
        ),
        migrations.AddField(
            model_name="riskincident",
            name="severity",
            field=models.ForeignKey(
                default=None,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="%(class)s",
                to="wbcompliance.riskincidenttype",
                verbose_name="Severity",
            ),
        ),
        migrations.AlterField(
            model_name="rulethreshold",
            name="severity",
            field=models.ForeignKey(
                default=None,
                help_text="The Triggered Severity when the rule is within the threshold range",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="thresholds",
                to="wbcompliance.riskincidenttype",
                verbose_name="Triggered Severity",
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="triggererincidentrelationship",
            name="severity",
            field=models.ForeignKey(
                default=None,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="%(class)s",
                to="wbcompliance.riskincidenttype",
                verbose_name="Severity",
            ),
        ),
        migrations.AlterField(
            model_name="complianceform",
            name="assigned_to",
            field=models.ManyToManyField(
                blank=True,
                help_text="List of group to whom this Compliance Form applies. It is applied to all users if not specified",
                related_name="compliance_forms",
                to="auth.group",
                verbose_name="Group to which the Compliance Form applies",
            ),
        ),
        migrations.AddField(
            model_name="riskincident",
            name="ignore_duration",
            field=models.DurationField(
                blank=True,
                help_text="If set, will ignore the forthcoming incidents for the specified duration",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="riskincident",
            name="last_ignored_date",
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AlterModelOptions(
            name="complianceform",
            options={"verbose_name": "Compliance Form", "verbose_name_plural": "Compliance Forms"},
        ),
        migrations.AlterModelOptions(
            name="compliancetype",
            options={
                "permissions": [("admin_compliance", "Can administer Compliance")],
                "verbose_name": "Compliance Type",
                "verbose_name_plural": "Compliance Types",
            },
        ),
        migrations.RenameField(
            model_name="complianceform",
            old_name="only_send_to_internals",
            new_name="only_internal",
        ),
        migrations.AlterField(
            model_name="complianceform",
            name="only_internal",
            field=models.BooleanField(
                default=True, help_text="Send the Form only to internal users", verbose_name="Only internal"
            ),
        ),
        migrations.RenameField(
            model_name="complianceformsignature",
            old_name="personal_notes",
            new_name="remark",
        ),
        migrations.AlterField(
            model_name="complianceformsignature",
            name="remark",
            field=models.TextField(blank=True, default="", null=True, verbose_name="Remark"),
        ),
        migrations.RenameField(
            model_name="complianceformsignature",
            old_name="policy_text",
            new_name="policy",
        ),
        migrations.AlterField(
            model_name="complianceformsignature",
            name="policy",
            field=models.TextField(blank=True, default="", null=True, verbose_name="Policy"),
        ),
        migrations.RenameField(
            model_name="compliancetask",
            old_name="occurance",
            new_name="occurrence",
        ),
        migrations.RenameField(
            model_name="reviewcompliancetask",
            old_name="occurance",
            new_name="occurrence",
        ),
        migrations.AlterField(
            model_name="compliancetask",
            name="occurrence",
            field=models.CharField(
                choices=[("YEARLY", "Yearly"), ("QUARTERLY", "Quarterly"), ("MONTHLY", "Monthly"), ("NEVER", "Never")],
                default="MONTHLY",
                max_length=32,
                verbose_name="Occurrence",
            ),
        ),
        migrations.AlterField(
            model_name="reviewcompliancetask",
            name="occurrence",
            field=models.CharField(
                choices=[("YEARLY", "Yearly"), ("QUARTERLY", "Quarterly"), ("MONTHLY", "Monthly"), ("NEVER", "Never")],
                default="MONTHLY",
                max_length=32,
                verbose_name="Occurrence",
            ),
        ),
        migrations.AlterField(
            model_name="complianceform",
            name="policy",
            field=models.TextField(blank=True, default="", null=True, verbose_name="Policy"),
        ),
        migrations.AlterField(
            model_name="complianceform",
            name="assigned_to",
            field=models.ManyToManyField(
                blank=True,
                related_name="forms_of_group",
                to="auth.group",
                verbose_name="Group to which the Form applies",
            ),
        ),
        migrations.AlterField(
            model_name="complianceform",
            name="changed",
            field=models.DateTimeField(auto_now=True, default=django.utils.timezone.now, verbose_name="Changed"),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="complianceformsignature",
            name="version",
            field=models.IntegerField(default=0, verbose_name="Version"),
        ),
        migrations.AlterField(
            model_name="reviewcompliancetask",
            name="changed",
            field=models.DateTimeField(auto_now=True, default=django.utils.timezone.now, verbose_name="Changed"),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="compliancetype",
            name="in_charge",
            field=models.ManyToManyField(
                blank=True,
                help_text="groups responsible for managing this type of compliance",
                related_name="compliance_types",
                to="auth.group",
                verbose_name="Group of administrators",
            ),
        ),
        migrations.AlterField(
            model_name="complianceformsignature",
            name="compliance_form",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="complianceforms",
                to="wbcompliance.complianceform",
                verbose_name="Form",
            ),
        ),
        migrations.AlterField(
            model_name="complianceformsignature",
            name="signed",
            field=models.DateTimeField(blank=True, null=True, verbose_name="Signed"),
        ),
        migrations.RenameField(
            model_name="complianceform",
            old_name="type",
            new_name="form_type",
        ),
        migrations.AlterField(
            model_name="complianceform",
            name="form_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                to="wbcompliance.complianceformtype",
                verbose_name="Form Type",
            ),
        ),
        migrations.AlterField(
            model_name="reviewcompliancetask",
            name="is_instance",
            field=models.BooleanField(default=False, verbose_name="Is occurrence"),
        ),
        migrations.AlterField(
            model_name="reviewcompliancetask",
            name="review_task",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="review_tasks",
                to="wbcompliance.reviewcompliancetask",
                verbose_name="Parent Review",
            ),
        ),
        migrations.AlterField(
            model_name="complianceformrule",
            name="ticked",
            field=models.BooleanField(default=False, verbose_name="Expected Answer"),
        ),
        migrations.AlterField(
            model_name="complianceformsignaturerule",
            name="ticked",
            field=models.BooleanField(default=False, verbose_name="Answer"),
        ),
        migrations.AlterField(
            model_name="compliancetype",
            name="name",
            field=models.CharField(max_length=255, unique=True, verbose_name="Name"),
        ),
        migrations.AlterField(
            model_name="complianceaction",
            name="description",
            field=models.TextField(blank=True, default="", null=True, verbose_name="Description"),
        ),
        migrations.AlterField(
            model_name="complianceaction",
            name="summary_description",
            field=models.TextField(blank=True, null=True, verbose_name="Summary Description"),
        ),
        migrations.AlterField(
            model_name="complianceevent",
            name="exec_summary",
            field=models.TextField(blank=True, null=True, verbose_name="Executive Summary"),
        ),
        migrations.AlterField(
            model_name="complianceevent",
            name="exec_summary_board",
            field=models.TextField(blank=True, null=True, verbose_name="Executive Summary for the Board"),
        ),
        migrations.AlterField(
            model_name="complianceevent",
            name="confidential",
            field=models.BooleanField(default=False, verbose_name="Confidential"),
        ),
        migrations.AlterField(
            model_name="complianceaction",
            name="active",
            field=models.BooleanField(default=True, verbose_name="Active"),
        ),
        migrations.AlterField(
            model_name="reviewcompliancetask",
            name="occured",
            field=models.DateField(blank=True, null=True, verbose_name="Occured Instance"),
        ),
        migrations.AlterField(
            model_name="compliancetaskinstance",
            name="occured",
            field=models.DateField(auto_now_add=True, verbose_name="Occured"),
        ),
        migrations.AlterField(
            model_name="compliancetask",
            name="review",
            field=models.ManyToManyField(
                blank=True,
                help_text="list of reviews that contain this task",
                related_name="tasks",
                to="wbcompliance.reviewcompliancetask",
                verbose_name="Review",
            ),
        ),
        migrations.AlterField(
            model_name="compliancetaskinstance",
            name="review",
            field=models.ManyToManyField(
                blank=True,
                help_text="list of reviews that contain this task instance",
                related_name="task_instances",
                to="wbcompliance.reviewcompliancetask",
                verbose_name="Review",
            ),
        ),
        migrations.AlterField(
            model_name="compliancetaskinstance",
            name="summary_text",
            field=models.TextField(blank=True, null=True, verbose_name="Summary Text"),
        ),
        migrations.AlterField(
            model_name="complianceaction",
            name="deadline",
            field=models.DateField(blank=True, null=True, verbose_name="Deadline"),
        ),
        migrations.AddField(
            model_name="complianceform",
            name="compliance_type",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="wbcompliance.compliancetype",
                verbose_name="Type",
            ),
        ),
        migrations.AddField(
            model_name="reviewcompliancetask",
            name="type",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="wbcompliance.compliancetype",
                verbose_name="Type",
            ),
        ),
        migrations.RunSQL(
            sql="SET CONSTRAINTS ALL IMMEDIATE;",
        ),
        migrations.AlterField(
            model_name="complianceform",
            name="compliance_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="wbcompliance.compliancetype", verbose_name="Type"
            ),
        ),
        migrations.AlterField(
            model_name="complianceaction",
            name="type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="actions_of_type",
                to="wbcompliance.compliancetype",
                verbose_name="Type",
            ),
        ),
        migrations.AlterField(
            model_name="complianceevent",
            name="type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="events_of_type",
                to="wbcompliance.compliancetype",
                verbose_name="Type",
            ),
        ),
        migrations.AlterField(
            model_name="compliancetask",
            name="type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="tasks_of_type",
                to="wbcompliance.compliancetype",
                verbose_name="Type",
            ),
        ),
        migrations.AlterField(
            model_name="reviewcompliancetask",
            name="type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="reviewtask_of_type",
                to="wbcompliance.compliancetype",
                verbose_name="Type",
            ),
        ),
        migrations.RunSQL(
            sql="SET CONSTRAINTS ALL DEFERRED;",
        ),
        migrations.RemoveIndex(
            model_name="ruletriggererrelationship",
            name="wbcomplianc_rule_id_c5fc0f_idx",
        ),
        migrations.RenameField(
            model_name="riskcheck",
            old_name="rule_triggerer_relationship",
            new_name="rule_checked_object_relationship",
        ),
        migrations.RenameField(
            model_name="rulebackend",
            old_name="allowed_triggerer_content_type",
            new_name="allowed_checked_object_content_type",
        ),
        migrations.RenameField(
            model_name="ruletriggererrelationship",
            old_name="triggerer_id",
            new_name="checked_object_id",
        ),
        migrations.RenameField(
            model_name="ruletriggererrelationship",
            old_name="triggerer_repr",
            new_name="checked_object_repr",
        ),
        migrations.RenameField(
            model_name="ruletriggererrelationship",
            old_name="triggerer_content_type",
            new_name="checked_object_content_type",
        ),
        migrations.AddField(
            model_name="riskcheck",
            name="trigger_content_type",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="triggered_checks",
                to="contenttypes.contenttype",
            ),
        ),
        migrations.AddField(
            model_name="riskcheck",
            name="trigger_id",
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="ruletriggererrelationship",
            name="checked_object_content_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="risk_management_checked_objects",
                to="contenttypes.contenttype",
            ),
        ),
        migrations.AlterField(
            model_name="ruletriggererrelationship",
            name="rule",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="checked_object_relationships",
                to="wbcompliance.riskrule",
            ),
        ),
        migrations.AlterField(
            model_name="triggererincidentrelationship",
            name="incident",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="checked_object_relationships",
                to="wbcompliance.riskincident",
            ),
        ),
        migrations.AddIndex(
            model_name="ruletriggererrelationship",
            index=models.Index(
                fields=["rule", "checked_object_content_type", "checked_object_id"],
                name="wbcomplianc_rule_id_18dca9_idx",
            ),
        ),
        migrations.RenameModel(
            old_name="TriggererIncidentRelationship",
            new_name="CheckedObjectIncidentRelationship",
        ),
        migrations.RenameModel(
            old_name="RuleTriggererRelationship",
            new_name="RuleCheckedObjectRelationship",
        ),
        migrations.AlterModelOptions(
            name="checkedobjectincidentrelationship",
            options={
                "verbose_name": "Incident to Checked Object relationship",
                "verbose_name_plural": "Incident to Checked Object relationships",
            },
        ),
        migrations.AlterModelOptions(
            name="rulecheckedobjectrelationship",
            options={
                "verbose_name": "Checked Object to Rule relationship",
                "verbose_name_plural": "Checked Object to Rule relationships",
            },
        ),
        migrations.RenameIndex(
            model_name="checkedobjectincidentrelationship",
            new_name="wbcomplianc_inciden_f3db32_idx",
            old_name="wbcomplianc_inciden_e01fa0_idx",
        ),
        migrations.RenameIndex(
            model_name="rulecheckedobjectrelationship",
            new_name="wbcomplianc_rule_id_f0f369_idx",
            old_name="wbcomplianc_rule_id_18dca9_idx",
        ),
        migrations.AlterField(
            model_name="riskcheck",
            name="rule_checked_object_relationship",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="checks",
                to="wbcompliance.rulecheckedobjectrelationship",
                verbose_name="Rule-Checked Object Relationship",
            ),
        ),
        migrations.AlterField(
            model_name="riskcheck",
            name="rule_checked_object_relationship",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="checks",
                to="wbcompliance.rulecheckedobjectrelationship",
                verbose_name="Rule-Checked Object Relationship",
            ),
        ),
        migrations.RenameField(
            model_name="riskrule",
            old_name="is_passive",
            new_name="only_passive_check_allowed",
        ),
        migrations.AddField(
            model_name="riskcheck",
            name="status",
            field=models.CharField(
                choices=[
                    ("PENDING", "Pending"),
                    ("RUNNING", "Running"),
                    ("FAILED", "Failed"),
                    ("SUCCESS", "Success"),
                    ("WARNING", "Warning"),
                ],
                default="PENDING",
                max_length=32,
                verbose_name="Status",
            ),
        ),
        migrations.AlterField(
            model_name="riskrule",
            name="only_passive_check_allowed",
            field=models.BooleanField(
                default=True,
                help_text="If False, This rule can only be triggered passively",
                verbose_name="Passive Only",
            ),
        ),
        migrations.AddField(
            model_name="checkedobjectincidentrelationship",
            name="resolved_by",
            field=models.ForeignKey(
                blank=True,
                help_text="The person that resolved or ignored this incident",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_handled",
                to="directory.person",
                verbose_name="Handled by",
            ),
        ),
        migrations.AddField(
            model_name="complianceaction",
            name="changer",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="updated_actions",
                to="directory.person",
                verbose_name="Changer",
            ),
        ),
        migrations.AddField(
            model_name="complianceaction",
            name="creator",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="compliance_actions",
                to="directory.person",
                verbose_name="Creator",
            ),
        ),
        migrations.AddField(
            model_name="complianceevent",
            name="changer",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="updated_events",
                to="directory.person",
                verbose_name="Changer",
            ),
        ),
        migrations.AddField(
            model_name="complianceevent",
            name="creator",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="compliance_events",
                to="directory.person",
                verbose_name="Creator",
            ),
        ),
        migrations.AddField(
            model_name="complianceform",
            name="changer",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="directory.person",
                verbose_name="Changer",
            ),
        ),
        migrations.AddField(
            model_name="complianceform",
            name="creator",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="compliance_forms",
                to="directory.person",
                verbose_name="Creator",
            ),
        ),
        migrations.AddField(
            model_name="complianceformsignature",
            name="person",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="signed_compliance_forms",
                to="directory.person",
                verbose_name="Signer",
            ),
        ),
        migrations.AddField(
            model_name="reviewcompliancetask",
            name="changer",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="directory.person",
                verbose_name="Changer",
            ),
        ),
        migrations.AddField(
            model_name="reviewcompliancetask",
            name="creator",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="author_review_compliane_tasks",
                to="directory.person",
                verbose_name="Creator",
            ),
        ),
        migrations.AddField(
            model_name="riskincident",
            name="resolved_by",
            field=models.ForeignKey(
                blank=True,
                help_text="The person that resolved or ignored this incident",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="%(class)s_handled",
                to="directory.person",
                verbose_name="Handled by",
            ),
        ),
        migrations.AddField(
            model_name="rulethreshold",
            name="notifiable_users",
            field=models.ManyToManyField(
                blank=True,
                help_text="Notified Persons for this rule and severity",
                related_name="notified_rule_thresholds",
                to="directory.person",
                verbose_name="Notifiable Persons",
            ),
        ),
        migrations.AlterField(
            model_name="checkedobjectincidentrelationship",
            name="severity",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="%(class)s",
                to="wbcompliance.riskincidenttype",
                verbose_name="Severity",
            ),
        ),
        migrations.AlterField(
            model_name="riskincident",
            name="severity",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="%(class)s",
                to="wbcompliance.riskincidenttype",
                verbose_name="Severity",
            ),
        ),
    ]



import numpy as np
import pandas as pd
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.utils.functional import cached_property
from django.utils.translation import gettext as _
from wbcore.pandas import fields as pf
from wbcore.pandas.views import PandasAPIViewSet
from wbcore.utils.models import get_object

from wbcompliance.filters import RiskManagementIncidentFilter
from wbcompliance.models.risk_management import (
    CheckedObjectIncidentRelationship,
    RiskRule,
)

from ..display import RiskManagementIncidentTableDisplayConfig
from ..endpoints import RiskManagementIncidentTableEndpointConfig
from ..titles import RiskManagementIncidentTableTitleConfig


class RiskManagementIncidentTableView(PandasAPIViewSet):
    IDENTIFIER = "wbcompliance:riskmanagementincidentpandas"
    filterset_class = RiskManagementIncidentFilter
    queryset = CheckedObjectIncidentRelationship.objects.all()

    display_config_class = RiskManagementIncidentTableDisplayConfig
    title_config_class = RiskManagementIncidentTableTitleConfig
    endpoint_config_class = RiskManagementIncidentTableEndpointConfig

    @cached_property
    def checked_object_content_type(self):
        if checked_object_content_type_id := self.request.GET.get("checked_object_content_type", None):
            return ContentType.objects.get(id=checked_object_content_type_id)

    @cached_property
    def get_rule_map(self):
        rules = RiskRule.objects.filter(
            Q(is_enable=True)
            & (
                Q(rule_backend__allowed_checked_object_content_type__isnull=True)
                | Q(rule_backend__allowed_checked_object_content_type=self.checked_object_content_type)
            )
        )
        return list(map(lambda x: (f'rule_{x["id"]}', x["name"]), rules.order_by("name").values("id", "name")))

    def get_ordering_fields(self):
        return ["checked_object_repr", *[x[0] for x in self.get_rule_map]]

    def get_pandas_fields(self, request):
        fields = [
            pf.PKField(key="id", label=_("ID")),
            pf.CharField(key="checked_object_repr", label=_("ID")),
        ]
        for key, label in self.get_rule_map:
            fields.append(pf.IntegerField(key=key, label=label))
        return pf.PandasFields(fields=tuple(fields))

    def get_queryset(self):
        return CheckedObjectIncidentRelationship.objects.filter(incident__rule__is_enable=True)

    def get_dataframe(self, request, queryset, **kwargs):
        df = pd.DataFrame()
        if checked_object_content_type := self.checked_object_content_type:
            df = pd.DataFrame(
                queryset.values(
                    "rule_check__checked_object_id",
                    "incident__rule__id",
                    "severity__severity_order",
                ),
                columns=[
                    "rule_check__checked_object_id",
                    "incident__rule__id",
                    "severity__severity_order",
                ],
            )
            df = df.rename(
                columns={
                    "rule_check__checked_object_id": "id",
                    "incident__rule__id": "rule",
                    "severity__severity_order": "severity",
                }
            )

            if not df.empty:
                df = (
                    df.pivot_table(index="id", columns=["rule"], values="severity", aggfunc=np.max, fill_value=-1)
                    .rename_axis(None, axis=1)
                    .astype("int")
                )
                df = df.rename(columns=lambda x: f"rule_{x}")
                df = df.reset_index()
                df["checked_object_repr"] = df.id.apply(
                    lambda x: str(get_object(checked_object_content_type.model_class(), x))
                )
                df = df.where(pd.notnull(df), -1)
        return df



import logging
from typing import TypeVar

from celery import shared_task
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.contrib.contenttypes.models import ContentType
from django.core.files.base import ContentFile
from django.db import models
from django.db.models import Q
from django.utils.translation import gettext_lazy as _
from slugify import slugify
from wbcore.contrib.documents.models import Document, DocumentType
from wbcore.contrib.documents.models.mixins import DocumentMixin
from wbcore.models import WBModel

User = get_user_model()
SelfComplianceType = TypeVar("SelfComplianceType", bound="ComplianceType")


def can_active_request(instance, user: "User") -> bool:
    if instance.changer:
        current_profile = instance.changer
    else:
        current_profile = instance.creator

    return user.is_superuser or (
        user.profile != current_profile and user.has_perm("wbcompliance.administrate_compliance")
    )


class ComplianceType(WBModel):
    class Meta:
        verbose_name = "Compliance Type"
        verbose_name_plural = "Compliance Types"
        permissions = [("administrate_compliance", "Can Administrate Compliance")]

    name = models.CharField(max_length=255, verbose_name=_("Name"), unique=True)
    description = models.TextField(default="", blank=True, verbose_name=_("Description"))
    in_charge = models.ManyToManyField(
        Group,
        related_name="compliance_types",
        blank=True,
        verbose_name=_("Group of administrators"),
        help_text=_("groups responsible for managing this type of compliance"),
    )

    def __str__(self) -> str:
        return "{}".format(self.name)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcompliance:compliancetype"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcompliance:compliancetyperepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{name}}"

    @classmethod
    def get_administrators(cls, type: SelfComplianceType | None = None) -> models.QuerySet["User"]:
        administrators = (
            get_user_model()
            .objects.filter(
                Q(groups__permissions__codename="administrate_compliance")
                | Q(user_permissions__codename="administrate_compliance")
            )
            .distinct()
        )
        if type:
            administrators = administrators.filter(groups__in=type.in_charge.all())
        return administrators

    @classmethod
    def is_administrator(cls, user) -> bool:
        return user.has_perm("wbcompliance.administrate_compliance")


@shared_task
def update_or_create_compliance_document(user_id: int, content_type_id: int, object_id: int, send_email: bool = True):
    user = get_user_model().objects.get(id=user_id)
    content_type = ContentType.objects.get(id=content_type_id)
    content_object = content_type.model_class().objects.get(id=object_id)

    filename = "{}.pdf".format(slugify(str(content_object)))
    logging.getLogger("weasyprint").setLevel(logging.CRITICAL)
    pdf_content = content_object.generate_pdf()
    logging.getLogger("weasyprint").setLevel(logging.INFO)
    content_file = ContentFile(pdf_content, name=filename)

    document_type, _ = DocumentType.objects.get_or_create(name="Compliance")
    document, _ = Document.objects.update_or_create(
        document_type=document_type,
        system_created=True,
        system_key="{}-{}-{}".format(content_type.model, content_object.id, filename),
        defaults={
            "file": content_file,
            "name": filename,
            "permission_type": Document.PermissionType.PRIVATE,
            "creator": user,
        },
    )
    document.link(content_object)
    if send_email:
        document.send_email(
            to_emails=user.email,
            as_link=True,
            subject="Compliance PDF - {}".format(str(content_object)),
        )


class ComplianceDocumentMixin(DocumentMixin):
    def get_permissions_for_user_and_document(self, user, view, created):
        """
        allows user to access the document associated with this object
        :return:     The tuple list permission for the corresponding user

        :note: Core implements a signal that automatically calls this function when instantiating the view
        """
        if ComplianceType.is_administrator(user):
            return [
                ("document.view_document", False),
                ("document.change_document", False),
                ("document.delete_document", False),
            ]
        return []



from typing import Optional

from django.utils.translation import gettext as _
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class ComplianceTypeDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="name", label=_("Name")),
                dp.Field(key="in_charge", label=_("Group of administrators")),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display([["name", "description", "in_charge", "administrators"]])



from reversion.views import RevisionMixin
from wbcore import viewsets

from wbcompliance.filters import ComplianceTypeFilter
from wbcompliance.models import ComplianceType
from wbcompliance.serializers import (
    ComplianceTypeModelSerializer,
    ComplianceTypeRepresentationSerializer,
)

from .display import ComplianceTypeDisplayConfig
from .endpoints import ComplianceTypeEndpointConfig
from .titles import ComplianceTypeTitleConfig


class ComplianceTypeRepresentationViewSet(viewsets.RepresentationViewSet):
    IDENTIFIER = "wbcompliance:compliancetyperepresentation"
    search_fields = ["name"]
    ordering_fields = ["name"]

    queryset = ComplianceType.objects.all()
    serializer_class = ComplianceTypeRepresentationSerializer


class ComplianceTypeModelViewSet(RevisionMixin, viewsets.ModelViewSet):
    IDENTIFIER = "wbcompliance:compliancetype"
    display_config_class = ComplianceTypeDisplayConfig
    endpoint_config_class = ComplianceTypeEndpointConfig
    title_config_class = ComplianceTypeTitleConfig

    search_fields = ["name"]
    ordering_fields = ["name", "in_charge"]
    ordering = ["name"]

    filterset_class = ComplianceTypeFilter

    serializer_class = ComplianceTypeModelSerializer

    queryset = ComplianceType.objects.prefetch_related("in_charge")
